# -*- coding: utf-8 -*-
"""
Created on Sun Jul  3 13:25:23 2022

@author: stefa
"""

import os
import sys
import shutil

this_file_dir = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.abspath(os.path.join(this_file_dir))
LIBRARY_DIR = os.path.join(PACKAGE_DIR, 'library')
exe_http = os.path.join(PACKAGE_DIR, 'bin', 'Hamilton HSLHttp Library Installer Version 2.7.exe')
exe_json = os.path.join(PACKAGE_DIR, 'bin', 'HSLJson Library v1.6 Installer.exe')

def full_paths_list(directory_abs_path):
    list_files = os.listdir(directory_abs_path)
    list_file_paths = [directory_abs_path + '\\' + file for file in list_files]
    return list_file_paths


def pyhamiltonconfig():
    print("Automatically configuring your PyHamilton installation")
    os.startfile(exe_http)
    os.startfile(exe_json)
    
    hamilton_lib_dir = os.path.abspath('C:/Program Files (x86)/HAMILTON/Library')

    
    def recursive_copy(source_dir, target_dir):
        source_list = full_paths_list(source_dir)
        for file in source_list:
            if os.path.isfile(file):
                shutil.copy(file, target_dir + '//' + os.path.basename(file))
            if os.path.isdir(file):
                target = target_dir + '//' + os.path.basename(file)
                if not os.path.exists(target):
                    os.mkdir(target)
                recursive_copy(source_dir + '//' + os.path.basename(file), target)            
    
    recursive_copy(LIBRARY_DIR, hamilton_lib_dir)        
    print("Configuration completed")


            
            
            
            


"""
Built-in commands, definitions of their parameters, and defaults.
"""

_channel_patt_16 = '1'*8 + '0'*8
_channel_patt_96 = '1'*96

_fan_port = 6
# on module load, scan COM ports to see if the usual fan COM number (6) has been reassigned by the OS
try:
    import serial.tools.list_ports
    for port in serial.tools.list_ports.comports():
        port_parse = str(port).split(' ')
        if 'Isolated' in port_parse and 'RS-485' in port_parse:
            _fan_port = int(port_parse[0][-1])
except Exception:
    pass

defaults_by_cmd = { # 'field':None indicates field is required when assembling command

    'initialize':('INITIALIZE', {
        'initializeAlways':0
    }),

    'channelTipPickUp':('PICKUP', {
        'tipSequence':'', # (string) leave empty if you are going to provide specific labwarePositions below
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above.'LabwareId1, positionId1; LabwareId2,positionId2; ....'
        'channelVariable':_channel_patt_16, # (string)  channel pattern e.g. '11110000'
        'sequenceCounting':0, # (integer) 0=don´t autoincrement,  1=Autoincrement
        'channelUse':1 # (integer) 1=use all sequence positions (no empty wells), 2=keep channel pattern
    }),

    'channelTipEject':('EJECT', {
        'wasteSequence':'', # (string) leave empty if you are going to provide specific labware-positions below or ejecting to default waste
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above.'LabwareId1, positionId1; LabwareId2,positionId2; ....'
        'channelVariable':_channel_patt_16, # (string) channel pattern e.g. "11110000"
        'sequenceCounting':0, # (integer) 0=don´t autoincrement,  1=Autoincrement.  Value omitted if ejecting to default waste
        'channelUse':1, # (integer) 1=use all sequence positions (no empty wells), 2=keep channel pattern
        'useDefaultWaste':0 # (integer) 0=eject to custom waste sequence,  1=Use default waste
    }),

    'channelAspirate':('ASPIRATE', {
        'aspirateSequence':'', # (string) leave empty if you are going to provide specific labware-positions below
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....'
        'volumes':None, # (float or string) enter a single value used for all channels or enter an array of values for each channel like [10.0,15.5,11.2]
        'channelVariable':_channel_patt_16, # (string) channel pattern e.g. "11110000"
        'liquidClass':None, # (string)
        'sequenceCounting':0, # (integer) 0=don´t autoincrement,  1=Autoincrement
        'channelUse':1, # (integer) 1=use all sequence positions (no empty wells), 2=keep channel pattern
        'aspirateMode':0, # (integer) 0=Normal Aspiration, 1=Consecutive (don´t aspirate blowout), 2=Aspirate all 
        'capacitiveLLD':0, # (integer) 0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition
        'pressureLLD':0, # (integer) 0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From liquid class definition
        'liquidFollowing':0, # (integer) 0=Off , 1=On
        'submergeDepth':2.0, # (float) mm of immersion below liquid´s surface to start aspiration when using LLD
        'liquidHeight':1.0, # (float) mm above container´s bottom to start aspiration when not using LLD
        'maxLLdDifference':0.0, # (float) max mm height different between cLLD and pLLD detected liquid levels
        'mixCycles':0, # (integer) number of mixing cycles (1 cycle = 1 asp + 1 disp)
        'mixPosition':0.0, # (float) additional immersion mm below aspiration position to start mixing
        'mixVolume':0.0, # (float) mix volume
        'xDisplacement':0.0,
        'yDisplacement':0.0,
        'zDisplacement':0.0,
        'airTransportRetractDist':10.0, # (float) mm to move up in Z after finishing the aspiration at a fixed height before aspirating 'transport air'
        'touchOff':0, # (integer) 0=Off , 1=On
        'aspPosAboveTouch':0.0 # (float)  mm to move up in Z after touch off detects the bottom before aspirating liquid
    }),

    'channelDispense':('DISPENSE', {
        'dispenseSequence':'', # (string) leave empty if you are going to provide specific labware-positions below
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....'
        'volumes':None, # (float or string) enter a single value used for all channels or enter an array of values for each channel like [10.0,15.5,11.2]
        'channelVariable':_channel_patt_16, # (string) channel pattern e.g. "11110000"
        'liquidClass':None, # (string)
        'sequenceCounting':0, # (integer) 0=don´t autoincrement,  1=Autoincrement
        'channelUse':1, # (integer) 1=use all sequence positions (no empty wells), 2=keep channel pattern
        'dispenseMode':8, # (integer) 0=Jet Part, 1=Jet Empty, 2=Surface Part, 3=Surface Empty, 4=Jet Drain tip, 8=From liquid class, 9=Blowout tip
        'capacitiveLLD':0, # (integer) 0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition
        'liquidFollowing':0, # (integer) 0=Off , 1=On
        'submergeDepth':2.0, # (float) mm of immersion below liquid´s surface to start dispense when using LLD
        'liquidHeight':1.0, # (float) mm above container´s bottom to start dispense when not using LLD
        'mixCycles':0, # (integer) number of mixing cycles (1 cycle = 1 asp + 1 disp)
        'mixPosition':0.0, # (float) additional immersion mm below dispense position to start mixing
        'mixVolume':0.0, # (float) mix volume
        'xDisplacement':0.0,
        'yDisplacement':0.0,
        'zDisplacement':0.0,
        'airTransportRetractDist':10.0, # (float) mm to move up in Z after finishing the dispense at a fixed height before aspirating 'transport air'
        'touchOff':0, # (integer) 0=Off , 1=On
        'dispPositionAboveTouch':0.0, # (float) mm to move up in Z after touch off detects the bottom, before dispense
        'zMoveAfterStep':0, # (integer) 0=normal, 1=Minimized (Attention!!! this depends on labware clearance height, can crash). 
        'sideTouch':0 # (integer) 0=Off , 1=On
    }),

    'mph96TipPickUp':('PICKUP96', {
        'tipSequence':'', # (string) leave empty if you are going to provide specific labware-positions below
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....' Must contain 96 values
        'channelVariable':_channel_patt_96, # (string) channel Variable e.g. "11110000...." . Must contain 96 values
        'sequenceCounting':0, # (integer) 0=don´t autoincrement,  1=Autoincrement
        'reducedPatternMode':0 # (integer) 0=All (not reduced), 1=One channel, 2=One row  3=One column
    }),

    'mph96TipEject':('EJECT96', {
        'wasteSequence':'', # (string) leave empty if you are going to provide specific labware-positions below or ejecting to default waste
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....'
        'channelVariable':_channel_patt_96, # (string) channel Variable e.g. "11110000...." . Must contain 96 values
        'sequenceCounting':0, # (integer)  0=don´t autoincrement,  1=Autoincrement.  Value omitted if ejecting to default waste
        'tipEjectToKnownPosition':0 # (integer) 0=Eject to specified sequence position,  1=Eject on tip pick up position, 2=Eject on default waste
    }),

    'mph96Aspirate':('ASPIRATE96', {
        'aspirateSequence':'', # (string) leave empty if you are going to provide specific labware-positions below
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above. LabwareId1, positionId1; LabwareId2,positionId2; ....
        'aspirateVolume':None, # (float)  single volume used for all channels in the head. There´s no individual control of each channel volume in multi-probe heads.
        'channelVariable':_channel_patt_96, # (string) channel Variable e.g. "11110000...." . Must contain 96 values
        'liquidClass':None, # (string)
        'sequenceCounting':0, # (integer)  0=don´t autoincrement,  1=Autoincrement
        'aspirateMode':0, # (integer) 0=Normal Aspiration, 1=Consecutive (don´t aspirate blowout), 2=Aspirate all 
        'capacitiveLLD':0, # (integer) 0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition
        'liquidFollowing':0, # (integer) 0=Off , 1=On
        'submergeDepth':2.0, # (float) mm of immersion below liquid´s surface to start aspiration when using LLD
        'liquidHeight':1.0, # (float) mm above container´s bottom to start aspiration when not using LLD
        'mixCycles':0, # (integer) number of mixing cycles (1 cycle = 1 asp + 1 disp)
        'mixPosition':0.0, # (float) additional immersion mm below aspiration position to start mixing
        'mixVolume':0.0, # (float) mix volume
        'airTransportRetractDist':10.0 # (float) mm to move up in Z after finishing the aspiration at a fixed height before aspirating 'transport air'
    }),

    'mph96Dispense':('DISPENSE96', {
        'dispenseSequence':'', # (string) leave empty if you are going to provide specific labware-positions below
        'labwarePositions':'', # (string) leave empty if you are going to provide a sequence name above. LabwareId1, positionId1; LabwareId2,positionId2; ....
        'dispenseVolume':None, # (float) single volume used for all channels in the head. There´s no individual control of each channel volume in multi-probe heads.
        'channelVariable':_channel_patt_96, # (string) channel Variable e.g. "11110000...." . Must contain 96 values
        'liquidClass':None, # (string) 
        'sequenceCounting':0, # (integer)  0=don´t autoincrement,  1=Autoincrement
        'dispenseMode':8, # (integer) 0=Jet Part, 1=Jet Empty, 2=Surface Part, 3=Surface Empty,4=Jet Drain tip, 8=From liquid class, 9=Blowout tip
        'capacitiveLLD':0, # (integer) 0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition
        'liquidFollowing':0, # (integer)  0=Off , 1=On
        'submergeDepth':2.0, # (float) mm of immersion below liquid´s surface to start dispense when using LLD
        'liquidHeight':1.0, # (float) mm above container´s bottom to start dispense when not using LLD
        'mixCycles':0, # (integer)  number of mixing cycles (1 cycle = 1 asp + 1 disp)
        'mixPosition':0.0, # (float)  additional immersion mm below dispense position to start mixing
        'mixVolume':0.0, # (float)  mix volume
        'airTransportRetractDist':10.0, # (float) mm to move up in Z after finishing the dispense at a fixed height before aspirating 'transport air'
        'zMoveAfterStep':0, # (integer) 0=normal, 1=Minimized (Attention!!! this depends on labware clearance height, can crash). 
        'sideTouch':0 # (integer) 0=Off , 1=On
    }),

    'iSwapGet':('ISWAP_GET', {
        'plateSequence':'', # leave empty if you are going to provide specific plate labware-position below
        'plateLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'lidSequence':'', # leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste
        'lidLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'toolSequence':'', # sequence name of the iSWAP. leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1;
        'sequenceCounting':0, # (integer) 0=don´t autoincrement plate sequence,  1=Autoincrement
        'movementType':0, # (integer) 0=To carrier, 1=Complex movement
        'transportMode':0, # (integer) 0=Plate only, 1=Lid only ,2=Plate with lid
        'gripForce':4, # (integer) 2 (minimum) ... 9 (maximum)
        'inverseGrip':0, # (integer) 0=Off, 1=On
        'collisionControl':0, # (integer) 0=Off, 1=On
        'gripMode':1, # (integer) 0=Small side, 1=Large side
        'retractDistance':0.0, # (float) retract distance [mm] (only used if 'movement type' is set to 'complex movement')
        'liftUpHeight':20.0, # (float) lift-up distance [mm] (only used if 'movement type' is set to 'complex movement')
        'gripWidth':123.7, # (float) grip width when closed [mm]
        'tolerance':2.0, # (float) tolerance [mm]
        'gripHeight':3.0, # (float) height to grip above bottom of labware [mm]
        'widthBefore':130.0, # (float) grip width when opened before grip [mm]
        'labwareOrientation':1
    }),

    'iSwapPlace':('ISWAP_PLACE', {
        'plateSequence':'', # leave empty if you are going to provide specific plate labware-position below
        'plateLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'lidSequence':'', # leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste
        'lidLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'toolSequence':'', # sequence name of the iSWAP. leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1;
        'sequenceCounting':0, # (integer) 0=don´t autoincrement plate sequence,  1=Autoincrement
        'movementType':0, # (integer) 0=To carrier, 1=Complex movement
        'transportMode':0, # (integer) 0=Plate only, 1=Lid only ,2=Plate with lid
        'collisionControl':0, # (integer) 0=Off, 1=On
        'retractDistance':0.0, # (float) retract distance [mm] (only used if 'movement type' is set to 'complex movement')
        'liftUpHeight':20.0 # (float) lift-up distance [mm] (only used if 'movement type' is set to 'complex movement')
    }),
    'iSwapMove':('ISWAP_MOVE',{
        'plateSequence':'',
        'plateLabwarePositions':'',
        'collisionControl':0,
        'gripMode':1
    }),

    'HxFanSet':('HEPA', {
        'deviceNumber':_fan_port, # (integer) COM port number of fan
        'persistant':1, # (integer) 0=don´t keep fan running after method exits, 1=keep settings after method exits
        'fanSpeed':None, # (float) set percent of maximum fan speed
        'simulate':0 #(integer) 0=normal mode, 1=use HxFan simulation mode
    }),

    'CORE96WashEmpty':('WASH96_EMPTY', {
        'refillAfterEmpty':0, # (integer) 0=Don't refill, 1=Refill both chambers, 2=Refill chamber 1 only, 3=Refill chamber 2 only
        'chamber1WashLiquid':0, # (integer) 0=Liquid 1 (red container), 1=liquid 2 (blue container)
        'chamber1LiquidChange':0, # (integer) 0=No, 1=Yes TODO: What does this mean?
        'chamber2WashLiquid':0, # (integer) 0=Liquid 1 (red container), 1=liquid 2 (blue container)
        'chamber2LiquidChange':0, # (integer) 0=No, 1=Yes TODO: What does this mean?
    }),

    'gripGet':('GRIP_GET', {
        'plateSequence':'', # leave empty if you are going to provide specific plate labware-position below
        'plateLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'lidSequence':'', # leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste
        'lidLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'toolSequence':'COREGripTool', # sequence name of the CO-RE Gripper
        'gripForce':3, # (integer) 0-9, from lowest to highest
        'gripperToolChannel':8, # specifies the higher of two consecutive integers representing the CO-RE gripper channels.
        'sequenceCounting':0, # (integer) 0=don´t autoincrement plate sequence,  1=Autoincrement
        'gripWidth':75, # (float) mm
        'gripHeight':3.0, # (float) mm
        'widthBefore':90, # (float) mm width before gripping
        'gripSpeed':5.0, # (float) mm/s. Must be supplied
        'zSpeed':50.0, # (float) mm/s. Must be supplied
        'transportMode':0, # (integer) 0=Plate only, 1=Lid only ,2=Plate with lid
        'checkPlate':0 # (integer) 
    }),

    'gripMove':('GRIP_MOVE', {
        'plateSequence':'', # leave empty if you are going to provide specific plate labware-position below
        'xAcceleration':4, # (integer) 1-5 from slowest to fastest, where 4 is default
        'plateLabwarePositions':'', # leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste

    }),

    'gripPlace':('GRIP_PLACE', {
        'plateSequence':'', # leave empty if you are going to provide specific plate labware-position below
        'plateLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'lidSequence':'', # leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste
        'lidLabwarePositions':'', # leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 
        'toolSequence':'COREGripTool', # sequence name of the iSWAP. leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1;
        'sequenceCounting':0, # (integer) 0=don´t autoincrement plate sequence,  1=Autoincrement
        'movementType':0, # (integer) 0=To carrier, 1=Complex movement
        'transportMode':0, # (integer) 0=Plate only, 1=Lid only ,2=Plate with lid
        'ejectToolWhenFinish':1, # (integer) 0=Off, 1=On
        'zSpeed':100.0, # (float) mm/s
        'platePressOnDistance':0.0, # (float) lift-up distance [mm] (only used if 'movement type' is set to 'complex movement'),
        'xAcceleration':4  # (integer) 1-5 from slowest to fastest, where 4 is default
    }),
    'moveSequence':('MOVE_SEQ',{

        'inputSequence':'',
        'xDisplacement':'',
        'yDisplacement':'',
        'zDisplacement':'',
    }),
    'TEC_Initialize':('TEC_INIT', {

        'ControllerID':'', # (integer)
        'SimulationMode':False, # 0=False, 1=True; 
    }),
    'TEC_StartTempControl':('TEC_START', {

        'ControllerID':'', # (integer)
        'DeviceID':'', # (integer); 
    }),

    'TEC_SetTarget':('TEC_SET_TARGET', {
        'ControllerID':'', # (integer)
        'DeviceID':'', # (integer); 
        'TargetTemperature':'', # (float); 
    }),
    'TEC_StopTemperatureControl':('TEC_STOP', {

        'ControllerID':'', # (integer)
        'DeviceID':'', # (integer); 
    }),
    'TEC_Terminate':('TEC_TERMINATE', {

        'StopAllDevices':1, # 0=False, 1=True
    }),
    'TiltModule_Initialize':('TILT_INIT', {

        'ModuleName':'', # (string)
        'Comport':'', # (integer)
        'TraceLevel':'', # (integer)
        'Simulate': '' # 0=False, 1=True
    }),
    'TiltModule_MoveToPosition':('TILT_MOVE', {

        'ModuleName':'', # (string)
        'Angle':'' # (integer)
    }),
    'FirmwareCommand':('FIRMWARECOMMAND', {

        'FirmwareCommandList':[], # list elements as {FirmwareCommand:'', FirmwareParameter:''} 
    }),
    'BarcodeReader_Initialize':('BC_INITIALIZE',{

        'ComPort':'' # (string)
    }),
    'BarcodeReader_Read':('BC_READ',{

    }),
}

"""All of the command names supported out of the box, mapped to their default params.

On module load, defaults_by_cmd is parsed into `HamiltonCmdTemplate`s, which are injected into the global package namespace under the first element of the values of this dict (strings in all caps). This is so that they can be imported directly from `pyhamilton` as module-level variables, while avoiding circular imports.

Example:

```
from pyhamilton import INITIALIZE
```


INITIALIZE

- initializeAlways (int)

    0=only initialize components not already initialized, 1=always reinitialize all robot components

    Default: 0 





PICKUP

- tipSequence (string)

    leave empty if you are going to provide specific labwarePositions below

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above.'LabwareId1, positionId1; LabwareId2,positionId2; ....'

    Default: ''

- channelVariable (string)

    channel pattern e.g. '11110000'

    Default: _channel_patt_16

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement

    Default: 0

- channelUse (integer)

    1=use all sequence positions (no empty wells), 2=keep channel pattern

    Default: 1



EJECT

- wasteSequence (string)

    leave empty if you are going to provide specific labware-positions below or ejecting to default waste

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above.'LabwareId1, positionId1; LabwareId2,positionId2; ....'

    Default: ''

- channelVariable (string)

    channel pattern e.g. "11110000"

    Default: _channel_patt_16

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement.  Value omitted if ejecting to default waste

    Default: 0

- channelUse (integer)

    1=use all sequence positions (no empty wells), 2=keep channel pattern

    Default: 1

- useDefaultWaste (integer)

    0=eject to custom waste sequence,  1=Use default waste

    Default: 0





ASPIRATE

- aspirateSequence (string)

    leave empty if you are going to provide specific labware-positions below

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....'

    Default: ''

- volumes (float or string)

    enter a single value used for all channels or enter an array of values for each channel like [10.0,15.5,11.2]

    Default: None

- channelVariable (string)

    channel pattern e.g. "11110000"

    Default: _channel_patt_16

- liquidClass (string)

    Default: None

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement

    Default: 0

- channelUse (integer)

    1=use all sequence positions (no empty wells), 2=keep channel pattern

    Default: 1

- aspirateMode (integer)

    0=Normal Aspiration, 1=Consecutive (don´t aspirate blowout), 2=Aspirate all 

    Default: 0

- capacitiveLLD (integer)

    0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition

    Default: 0

- pressureLLD (integer)

    0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From liquid class definition

    Default: 0

- liquidFollowing (integer)

    0=Off , 1=On

    Default: 0

- submergeDepth (float)

    mm of immersion below liquid´s surface to start aspiration when using LLD

    Default: 2.0

- liquidHeight (float)

    mm above container´s bottom to start aspiration when not using LLD

    Default: 1.0

- maxLLdDifference (float)

    max mm height different between cLLD and pLLD detected liquid levels

    Default: 0.0

- mixCycles (integer)

    number of mixing cycles (1 cycle = 1 asp + 1 disp)

    Default: 0

- mixPosition (float)

    additional immersion mm below aspiration position to start mixing

    Default: 0.0

- mixVolume (float)

    mix volume

    Default: 0.0

- airTransportRetractDist (float)

    mm to move up in Z after finishing the aspiration at a fixed height before aspirating 'transport air'

    Default: 10.0

- touchOff (integer)

    0=Off , 1=On

    Default: 0

- aspPosAboveTouch (float)

    mm to move up in Z after touch off detects the bottom before aspirating liquid

    Default: 0.0





DISPENSE

- dispenseSequence (string)

    leave empty if you are going to provide specific labware-positions below

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....'

    Default: ''

- volumes (float or string)

    enter a single value used for all channels or enter an array of values for each channel like [10.0,15.5,11.2]

    Default: None

- channelVariable (string)

    channel pattern e.g. "11110000"

    Default: _channel_patt_16

- liquidClass (string)

    Default: None

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement

    Default: 0

- channelUse (integer)

    1=use all sequence positions (no empty wells), 2=keep channel pattern

    Default: 1

- dispenseMode (integer)

    0=Jet Part, 1=Jet Empty, 2=Surface Part, 3=Surface Empty, 4=Jet Drain tip, 8=From liquid class, 9=Blowout tip

    Default: 8

- capacitiveLLD (integer)

    0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition

    Default: 0

- liquidFollowing (integer)

    0=Off , 1=On

    Default: 0

- submergeDepth (float)

    mm of immersion below liquid´s surface to start dispense when using LLD

    Default: 2.0



- liquidHeight (float)

    mm above container´s bottom to start dispense when not using LLD

    Default: 1.0

- mixCycles (integer)

    number of mixing cycles (1 cycle = 1 asp + 1 disp)

    Default: 0

- mixPosition (float)

    additional immersion mm below dispense position to start mixing

    Default: 0.0

- mixVolume (float)

    mix volume

    Default: 0.0

- airTransportRetractDist (float)

    mm to move up in Z after finishing the dispense at a fixed height before aspirating 'transport air'

    Default: 10.0

- touchOff (integer)

    0=Off , 1=On

    Default: 0

- dispPositionAboveTouch (float)

    mm to move up in Z after touch off detects the bottom, before dispense

    Default: 0.0

- zMoveAfterStep (integer)

    0=normal, 1=Minimized (Attention!!! this depends on labware clearance height, can crash). 

    Default: 0

- sideTouch (integer)

    0=Off , 1=On

    Default: 0





PICKUP96

- tipSequence (string)

    leave empty if you are going to provide specific labware-positions below

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....' Must contain 96 values

    Default: ''

- channelVariable (string)

    channel Variable e.g. "11110000...." . Must contain 96 values

    Default: _channel_patt_96

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement

    Default: 0

- reducedPatternMode (integer)

    0=All (not reduced), 1=One channel, 2=One row  3=One column

    Default: 0





EJECT96

- wasteSequence (string)

    leave empty if you are going to provide specific labware-positions below or ejecting to default waste

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above. 'LabwareId1, positionId1; LabwareId2,positionId2; ....'

    Default: ''

- channelVariable (string)

    channel Variable e.g. "11110000...." . Must contain 96 values

    Default: _channel_patt_96

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement.  Value omitted if ejecting to default waste

    Default: 0

- tipEjectToKnownPosition (integer)

    0=Eject to specified sequence position,  1=Eject on tip pick up position, 2=Eject on default waste

    Default: 0





ASPIRATE96

- aspirateSequence (string)

    leave empty if you are going to provide specific labware-positions below

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above. LabwareId1, positionId1; LabwareId2,positionId2; ....

    Default: ''

- aspirateVolume (float)

    single volume used for all channels in the head. There´s no individual control of each channel volume in multi-probe heads.

    Default: None

- channelVariable (string)

    channel Variable e.g. "11110000...." . Must contain 96 values

    Default: _channel_patt_96

- liquidClass (string)

    Default: None

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement

    Default: 0

- aspirateMode (integer)

    0=Normal Aspiration, 1=Consecutive (don´t aspirate blowout), 2=Aspirate all 

    Default: 0

- capacitiveLLD (integer)

    0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition

    Default: 0

- liquidFollowing (integer)

    0=Off , 1=On

    Default: 0

- submergeDepth (float)

    mm of immersion below liquid´s surface to start aspiration when using LLD

    Default: 2.0

- liquidHeight (float)

    mm above container´s bottom to start aspiration when not using LLD

    Default: 1.0

- mixCycles (integer)

    number of mixing cycles (1 cycle = 1 asp + 1 disp)

    Default: 0

- mixPosition (float)

    additional immersion mm below aspiration position to start mixing

    Default: 0.0

- mixVolume (float)

    mix volume

    Default: 0.0

- airTransportRetractDist (float)

    mm to move up in Z after finishing the aspiration at a fixed height before aspirating 'transport air'

    Default: 10.0





DISPENSE96

- dispenseSequence (string)

    leave empty if you are going to provide specific labware-positions below

    Default: ''

- labwarePositions (string)

    leave empty if you are going to provide a sequence name above. LabwareId1, positionId1; LabwareId2,positionId2; ....

    Default: ''

- dispenseVolume (float)

    single volume used for all channels in the head. There´s no individual control of each channel volume in multi-probe heads.

    Default: None

- channelVariable (string)

    channel Variable e.g. "11110000...." . Must contain 96 values

    Default: _channel_patt_96

- liquidClass (string)

    Default: None

- sequenceCounting (integer)

    0=don´t autoincrement,  1=Autoincrement

    Default: 0

- dispenseMode (integer)

    0=Jet Part, 1=Jet Empty, 2=Surface Part, 3=Surface Empty,4=Jet Drain tip, 8=From liquid class, 9=Blowout tip

    Default: 8

- capacitiveLLD (integer)

    0=Off, 1=Max, 2=High, 3=Mid, 4=Low, 5=From labware definition

    Default: 0

- liquidFollowing (integer)

    0=Off , 1=On

    Default: 0

- submergeDepth (float)

    mm of immersion below liquid´s surface to start dispense when using LLD

    Default: 2.0

- liquidHeight (float)

    mm above container´s bottom to start dispense when not using LLD

    Default: 1.0

- mixCycles (integer)

    number of mixing cycles (1 cycle = 1 asp + 1 disp)

    Default: 0

- mixPosition (float)

    additional immersion mm below dispense position to start mixing

    Default: 0.0

- mixVolume (float)

    mix volume

    Default: 0.0

- airTransportRetractDist (float)

    mm to move up in Z after finishing the dispense at a fixed height before aspirating 'transport air'

    Default: 10.0

- zMoveAfterStep (integer)

    0=normal, 1=Minimized (Attention!!! this depends on labware clearance height, can crash). 

    Default: 0

- sideTouch (integer)

    0=Off , 1=On

    Default: 0





ISWAP_GET

plateSequence

    leave empty if you are going to provide specific plate labware-position below

    Default:''

- plateLabwarePositions (string)

    leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 

    Default: ''

- lidSequence (string)

    leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste

    Default: ''

- lidLabwarePositions (string)

    leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 

    Default: ''

- toolSequence (string)

    sequence name of the iSWAP. leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1;

    Default: ''

- sequenceCounting (integer)

    0=don´t autoincrement plate sequence,  1=Autoincrement

    Default: 0

- movementType (integer)

    0=To carrier, 1=Complex movement

    Default: 0

- transportMode (integer)

    0=Plate only, 1=Lid only ,2=Plate with lid

    Default: 0

- gripForce (integer)

    2 (minimum) ... 9 (maximum)

    Default: 4

- inverseGrip (integer)

    0=Off, 1=On

    Default: 0

- collisionControl (integer)

    0=Off, 1=On

    Default: 0

- gripMode (integer)

    0=Small side, 1=Large side

    Default: 1

- retractDistance (float)

    retract distance [mm] (only used if 'movement type' is set to 'complex movement')

    Default: 0.0

- liftUpHeight (float)

    lift-up distance [mm] (only used if 'movement type' is set to 'complex movement')

    Default: 20.0

- gripWidth (float)

    grip width when closed [mm]

    Default: 123.7

- tolerance (float)

    tolerance [mm]

    Default: 2.0

- gripHeight (float)

    height to grip above bottom of labware [mm]

    Default: 3.0

- widthBefore (float)

    grip width when opened before grip [mm]

    Default: 130.0





ISWAP_PLACE

- plateSequence (string)

    leave empty if you are going to provide specific plate labware-position below

    Default: ''

- plateLabwarePositions (string)

    leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 

    Default: ''

- lidSequence (string)

    leave empty if you don´t use lid or if you are going to provide specific plate labware-positions below or ejecting to default waste

    Default: ''

- lidLabwarePositions (string)

    leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1; 

    Default: ''

- toolSequence (string)

    sequence name of the iSWAP. leave empty if you are going to provide a plate sequence name above. LabwareId1, positionId1;

    Default: ''

- sequenceCounting (integer)

    0=don´t autoincrement plate sequence,  1=Autoincrement

    Default: 0

- movementType (integer)

    0=To carrier, 1=Complex movement

    Default: 0

- transportMode (integer)

    0=Plate only, 1=Lid only ,2=Plate with lid

    Default: 0

- collisionControl (integer)

    0=Off, 1=On

    Default: 0

- retractDistance (float)

    retract distance [mm] (only used if 'movement type' is set to 'complex movement')

    Default: 0.0

- liftUpHeight (float)

    lift-up distance [mm] (only used if 'movement type' is set to 'complex movement')

    Default: 20.0





HEPA

- deviceNumber (integer)

    COM port number of fan

    Default: _fan_port

- persistant (integer)

    0=don´t keep fan running after method exits, 1=keep settings after method exits

    Default: 1

- fanSpeed (float)

    set percent of maximum fan speed

    Default: None

- simulate (integer)

    0=normal mode, 1=use HxFan simulation mode

    Default: 0 





WASH96_EMPTY

- refillAfterEmpty (integer)

    0=Don't refill, 1=Refill both chambers, 2=Refill chamber 1 only, 3=Refill chamber 2 only

    Default: 0

- chamber1WashLiquid (integer)

    0=Liquid 1 (red container), 1=liquid 2 (blue container)

    Default: 0

- chamber1LiquidChange (integer)

    0=No, 1=Yes TODO: What does this mean?

    Default: 0

- chamber2WashLiquid (integer)

    0=Liquid 1 (red container), 1=liquid 2 (blue container)

    Default: 0

- chamber2LiquidChange (integer)

    0=No, 1=Yes TODO: What does this mean?

    Default: 0


"""




"""
Couplings to Hamilton deck layouts.

Module `pyhamilton.deckresource` provides convenience classes and methods for interacting safely with Hamilton's Layout (`.lay`) files. It also implements transformations between well indexes and coordinates for a variety of labware, such as plates and tips.
"""
import string, shutil, os, string, re
from datetime import datetime
from pyhamilton import OEM_LAY_PATH, LAY_BACKUP_DIR
from .oemerr import ResourceUnavailableError


class ResourceType:
    """
    Specifies a type of labware to extract using LayoutManager, and how.

    This class associates a resource class, such as `Plate96`, with either a literal labware identifier (`str`) that appears in the Hamilton Layout (`.lay`) file, or a pair of functions: one that identifies when a text line in a layout file could be assigned this resource, called `test`, and one that parses such a name out of the line, called `extract_name`.

    Typical usage:

    ```
    plate_type = ResourceType(Plate96, 'Cos_96_Rd_0001')
    lmgr = LayoutManager('layout.lay')
    plate = lmgr.assign_unused_resource(plate_type)
    ```

    Or:

    ```
    plate_type = ResourceType(Plate96,
            LayoutManager.line_has_prefixed_name('Cos_96_Rd_'),
            LayoutManager.name_from_line)
    lmgr = LayoutManager('layout.lay')
    plate1 = lmgr.assign_unused_resource(plate_type)
    plate2 = lmgr.assign_unused_resource(plate_type)
    ```

    Args:
      resource_class (class): a class that inherits from `DeckResource`. Instances of this class will be returned from `LayoutManager` when assigning resources, factory-style.
      *name_specifiers (list): This argument is unpacked with the "splat" operator (`*`) to enable polymorphism. One or the other of:
      - (two-argument form) an exact name (`str`) of a labware item that appears in the target layout file, or
      - (three-argument form) `test` and `extract_name` (see usage above):
          * `test`: a function [(`str`) -> `bool`] that identifies Layout file lines (`str`) that could be used to assign resources of this type
          * `extract_name`: a function [(`str`) -> `str`] that gets the desired name out of a line identified with `test`.

    """

    def __init__(self, resource_class, *args):
        self.resource_class = resource_class
        self.not_found_msg = None
        try:
            specific_name, = args
            self.test = lambda line: specific_name in re.split(r'\W', line)
            self.extract_name = lambda line: specific_name
            self.not_found_msg = 'No exact match for name "' + specific_name + '" to assign a resource of type ' + resource_class.__name__
        except ValueError:
            self.test, self.extract_name = args


class LayoutManager:
    """Optionally activates a Hamilton layout and helps access its contents.

    A `LayoutManager` manages the consistent assignment of `DeckResource` objects to items in a Hamilton Layout file (`.lay`). A `LayoutManager` must be used to set the active pyhamilton layout file, but use of this class is strictly optional when sending `pyhamilton` commands using `send_command`; names may be passed as string literals in commands instead if they are known in advance. The advantage to specifying all labware using `ResourceManager` is that resource names are verified to be present in the active layout file at runtime, and guaranteed never used more than once, both of which are necessary to avoid silent Hamilton errors.

    Example usage:

    ```
    lmgr = LayoutManager('layout.lay')
    plate = lmgr.assign_unused_resource(ResourceType(Plate24, 'plate_0'))
    culture_reservoir = lmgr.assign_unused_resource(ResourceType(Plate96, 'culture'))
    inducer_tips = lmgr.assign_unused_resource(ResourceType(Tip96, 'inducer_tips'))
    ```
    """

    _managers = {}
    @staticmethod
    def get_manager(checksum):
        """Return a `LayoutManager` previously instantiated for a layout file that has the specified checksum.

        Typically used when accessing the same layout file from multiple "threads" in the same process (using the `threading` module) to prevent name double-counting.

        Args:
          checksum (str): a checksum found at the end of a Hamilton Layout (`.lay`) file.
        """
        return LayoutManager._managers[checksum]

    @staticmethod
    def initial_printable(line, start=0):
        if not line:
            return ''
        end = start
        while end < len(line) and line[end] in string.printable:
            end += 1
        return line[start:end]

    @staticmethod
    def layline_objid(line):
        keys = 'ObjId', 'LabwareName'
        if 'Labware' in LayoutManager.layline_first_field(line):
            keys = 'Id', *keys
        for key in keys:
            try:
                start = line.index(key) + len(key) + 1
                return LayoutManager.initial_printable(line, start)
            except ValueError:
                pass
        else:
            return None

    @staticmethod
    def layline_first_field(line):
        return LayoutManager.initial_printable(line)

    @staticmethod
    def field_starts_with(field, prefix):
        try:
            return field.index(prefix) == 0
        except ValueError:
            return False

    @staticmethod
    def name_from_line(line):
        field = LayoutManager.layline_objid(line)
        if field:
            return field
        return LayoutManager.layline_first_field(line)

    @staticmethod
    def line_has_prefixed_name(prefix):
        def has_prefix(line):
            return LayoutManager.field_starts_with(LayoutManager.name_from_line(line), prefix)
        return has_prefix

    @staticmethod
    def _read_layfile_lines(layfile_path):
        buff = ''
        lines = []
        with open(layfile_path, 'rb') as f:
            for c in f.read():
                try:
                    c = bytes([c]).decode('utf-8')
                except UnicodeDecodeError:
                    continue
                buff += c
                if c in '\n\r\t':
                    lines.append(buff.strip())
                    buff = ''
        if buff:
            lines.append(buff)
        return lines

    @staticmethod
    def _layfile_checksum(layfile_path):
        lay_lines = LayoutManager._read_layfile_lines(layfile_path)
        return lay_lines[-1].split('checksum=')[1].split('$$')[0]

    @staticmethod
    def layfiles_equal(lay_path_1, lay_path_2):
        return LayoutManager._layfile_checksum(lay_path_1) == LayoutManager._layfile_checksum(lay_path_2)

    def __init__(self, layfile_path, install=True):
        self.lines = self._read_layfile_lines(layfile_path)
        self.resources = {}
        self.checksum = self._layfile_checksum(layfile_path)
        self._managers[self.checksum] = self
        if install and not LayoutManager.layfiles_equal(layfile_path, OEM_LAY_PATH):
                print('BACKING UP AND INSTALLING NEW LAYFILE')
                shutil.copy2(layfile_path, os.path.join(LAY_BACKUP_DIR, datetime.today().strftime('%Y%m%d_%H%M%S_') + os.path.basename(layfile_path)))
                shutil.copy2(layfile_path, OEM_LAY_PATH)

    def assign_unused_resource(self, restype, order_key=None, reverse=False):
        """Create a new deck resource after finding and assigning an unused name that matches the resource type.

        This method searches through the layout file for one new layout name that matches the given resource type. It reserves this layout name permanently so that no later calls to `assign_unused_resource` can create a deck resource with the same layout name. Returns a `DeckResource`.

        Args:
          restype (ResourceType): The resource type, which consists of a resource class (descendent of `DeckResource`) and some string pattern matching functions to identify the desired layout names.
          order_key (Callable[[DeckResource], Comparable]): Optional; when multiple layout names match, specifies a function of one argument that is used to extract a comparison key from each candidate `DeckResource` object. The arg-min or arg-max of `order_key` will be returned, depending on `reverse`. By default, lexicographic order by layout name is used, which is suitable for most use cases, e.g. plates with layout names "pcr-plate-a", "pcr-plate-b", "pcr-plate-c", ... will be returned in the expected order.
          reverse (bool): Optional; use reverse-lexicographic order for layout names, useful for e.g. plate stacking applications, or reverse the order imposed by `order_key` if it is given.
        Returns:

          A new instance of the resource class (descendent of `DeckResource`) from the given `ResourceType` `restype`.

        Raises:
          ResourceUnavailableError: no names in the layout file that have not already been assigned match the resource type

        """
        if order_key is None:
            order_key = lambda r: r.layout_name()
        if not isinstance(restype, ResourceType):
            raise TypeError('Must provide a ResourceType to be assigned')
        matching_ress = []
        for line in self.lines:
            if restype.test(line):
                match_name = restype.extract_name(line)
                if match_name in self.resources:
                    continue
                matching_ress.append(restype.resource_class(match_name))
        if not matching_ress:
            msg = restype.not_found_msg or 'No unassigned resource of type ' + restype.resource_class.__name__ + ' available'
            raise ResourceUnavailableError(msg)
        choose = max if reverse else min
        new_res = choose(matching_ress, key=order_key)
        self.resources[new_res.layout_name()] = new_res
        return new_res


class ResourceIterItem:

    def __init__(self, resource, index):
        self.parent_resource = resource
        self.index = index
        self.history = []


class Tip(ResourceIterItem):
    pass


class Vessel(ResourceIterItem):

    ADD = 0
    REMOVE = 1

    def record_removal(self, ml, dest=None):
        if dest is not None and not isinstance(dest, Vessel):
            raise ValueError('Sources and destinations in Vessel contents records must be Vessels')
        self.history.append((Vessel.REMOVE, ml, dest))

    def record_addition(self, ml, source):
        if not isinstance(source, Vessel):
            raise ValueError('Sources and destinations in Vessel contents records must be Vessels')
        self.history.append((Vessel.ADD, ml, source))

    def current_volume(self):
        return sum((ml if direction == Vessel.ADD else -ml) for direction, ml, _ in self.history)


class DeckResource:

    class align:
        VERTICAL = 'v'
        STD_96 = 'std_96'

    class types:
        TIP, VESSEL = range(2)

    def __init__(self, layout_name):
        raise NotImplementedError()

    def _alignment_delta(self, int_start, int_end):
        raise NotImplementedError() # (delta x, delta y, alignment properties list)

    def _assert_idx_in_range(self, idx_or_vessel):
        if isinstance(idx_or_vessel, Vessel):
            idx = idx_or_vessel.index
        else:
            idx = idx_or_vessel
        if not 0 <= idx < self._num_items:
            raise ValueError('Index ' + str(idx) + ' not in range for resource')

    def layout_name(self):
        """The layout name of this specific deck resource.

        Returns:
          The name (`str`) associated with this specific deck resource in the Hamilton Layout (`.lay`) file it came from.
        """
        return self._layout_name # default; override if needed. (str) 

    def position_id(self, idx):
        """The identifier used for one of a sequence of positions inside this labware.

        For labware with multiple positions, each position has a different identity, usually represented as a short string that will match the identifier scheme for this resource in the Hamilton Layout file it came from. The identifiers will usually be familiar from a laboratory setting.

        Examples

        - 96-well plates have 96 positions, each identified with a letter and a number like `'D4'`. For a `Plate96` instance named `plate`, `plate.position_id(0)` is `'A1'`, `plate.position_id(1)` is `'B1'`, and `plate.position_id(95)` is `'H12'`.
        - Hamilton racks of 96 tips have 96 positions, identified with integer strings like `'87'` that start with `'1'` at the top left tip and increase down columns (8 positions each) first. For a `Tip96` instance named `tips`, `tips.position_id(0)` is `'1'`, `tips.position_id(1)` is `'2'`, and `tips.position_id(95)` is `'96'`.

        Args:
          idx (int): the index into the sequence of positions. Note: `idx` is zero-indexed across all labware according to python convention, while most real-world labware positions are 1-indexed.

        Returns:
          The identifier (`str`) associated with the position `idx` specific deck resource in the Hamilton Layout (`.lay`) file it came from.

        Raises:
          NotImplementedError: The deck resource does not have positions.
        """
        raise NotImplementedError()

    def alignment_delta(self, start, end):
        args = {'start':start, 'end':end}
        for pos in args:
            if isinstance(args[pos], Vessel):
                if args[pos].parent_resource is not self:
                    raise ValueError('Positions provided as vessels, '
                            'but do not belong to this resource')
                args[pos] = start.index
            else:
                try:
                    args[pos] += 0
                except TypeError:
                    raise ValueError('Positions provided for delta must be integers or vessels')
            self._assert_idx_in_range(args[pos])
        return self._alignment_delta(args['start'], args['end'])

    def __iter__(self):
        for i in self._items:
            yield i


class Standard96(DeckResource):
    """Labware types with 96 positions that use a letter-number id scheme like `'A1'`.
    """

    def well_coords(self, idx):
        self._assert_idx_in_range(idx)
        return int(idx)//8, int(idx)%8

    def _alignment_delta(self, start, end):
        [self._assert_idx_in_range(p) for p in (start, end)]
        xs, ys = self.well_coords(start)
        xe, ye = self.well_coords(end)
        return (xe - xs, ye - ys, [DeckResource.align.STD_96]
                                  + ([DeckResource.align.VERTICAL] if xs == xe and ys != ye else []))

    def position_id(self, idx):
        x, y = self.well_coords(idx)
        return 'ABCDEFGH'[y] + str(x + 1)


class Tip96(Standard96):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 96
        self.resource_type = DeckResource.types.TIP
        self._items = [Tip(self, i) for i in range(self._num_items)]

    def position_id(self, idx): # tips use 1-indexed int ids descending columns first
        self._assert_idx_in_range(idx)
        return str(idx + 1) # switch to standard advance through row first


class Plate96(Standard96):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 96
        self.resource_type = DeckResource.types.VESSEL
        self._items = [Vessel(self, i) for i in range(self._num_items)]


class Plate24(DeckResource):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 24
        self.resource_type = DeckResource.types.VESSEL
        self._items = [Vessel(self, i) for i in range(self._num_items)]

    def well_coords(self, idx):
        self._assert_idx_in_range(idx)
        return int(idx)//4, int(idx)%4

    def _alignment_delta(self, start, end):
        [self._assert_idx_in_range(p) for p in (start, end)]
        xs, ys = self.well_coords(start)
        xe, ye = self.well_coords(end)
        return (xe - xs, ye - ys, [DeckResource.align.VERTICAL] if xs == xe and ys != ye else [])

    def position_id(self, idx):
        x, y = self.well_coords(idx)
        return 'ABCD'[y] + str(x + 1)


class Plate12(DeckResource):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 12
        self.resource_type = DeckResource.types.VESSEL
        self._items = [Vessel(self, i) for i in range(self._num_items)]

    def well_coords(self, idx):
        self._assert_idx_in_range(idx)
        return int(idx)//3, int(idx)%3

    def _alignment_delta(self, start, end):
        [self._assert_idx_in_range(p) for p in (start, end)]
        xs, ys = self.well_coords(start)
        xe, ye = self.well_coords(end)
        return (xe - xs, ye - ys, [DeckResource.align.VERTICAL] if xs == xe and ys != ye else [])

    def position_id(self, idx):
        x, y = self.well_coords(idx)
        return 'ABC'[y] + str(x + 1)

class Plate6(DeckResource):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 6
        self.resource_type = DeckResource.types.VESSEL
        self._items = [Vessel(self, i) for i in range(self._num_items)]

    def well_coords(self, idx):
        self._assert_idx_in_range(idx)
        return int(idx)//2, int(idx)%2

    def _alignment_delta(self, start, end):
        [self._assert_idx_in_range(p) for p in (start, end)]
        xs, ys = self.well_coords(start)
        xe, ye = self.well_coords(end)
        return (xe - xs, ye - ys, [DeckResource.align.VERTICAL] if xs == xe and ys != ye else [])

    def position_id(self, idx):
        x, y = self.well_coords(idx)
        return 'AB'[y] + str(x + 1)


class Plate384(DeckResource):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 384
        self.resource_type = DeckResource.types.VESSEL
        self._items = [Vessel(self, i) for i in range(self._num_items)]

    def well_coords(self, idx):
        self._assert_idx_in_range(idx)
        return int(idx)//16, int(idx)%16

    def _alignment_delta(self, start, end):
        [self._assert_idx_in_range(p) for p in (start, end)]
        xs, ys = self.well_coords(start)
        xe, ye = self.well_coords(end)
        return (xe - xs, ye - ys, [DeckResource.align.VERTICAL] if xs == xe and ys != ye else [])

    def position_id(self, idx):
        x, y = self.well_coords(idx)
        return 'ABCDEFGHIJKLMNOP'[y] + str(x + 1)

class Plate1536(DeckResource):

    def __init__(self, layout_name):
        self._layout_name = layout_name
        self._num_items = 1536
        self.resource_type = DeckResource.types.VESSEL
        self._items = [Vessel(self, i) for i in range(self._num_items)]

    def well_coords(self, idx):
        self._assert_idx_in_range(idx)
        return int(idx)//32, int(idx)%32

    def _alignment_delta(self, start, end):
        [self._assert_idx_in_range(p) for p in (start, end)]
        xs, ys = self.well_coords(start)
        xe, ye = self.well_coords(end)
        return (xe - xs, ye - ys, [DeckResource.align.VERTICAL] if xs == xe and ys != ye else [])

    def position_id(self, idx):
        x, y = self.well_coords(idx)
        row_letters = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ') + ['AA', 'AB', 'AC', 'AD', 'AE', 'AF']
        return row_letters[y] + str(x + 1)



class ChannelHead:

    def __init__(self):
        pass

    def pack_moves(self, moves):
        pass # return (binary channel pattern string, remaining moves)

    def can_move_simul(move1, move2): # both arguments are 4-tuples, (tips, ml, source, dest) 
        tips, mls, sources, dests = zip(move1, move2)
        move_components = zip(sources, tips, dests)
        xy_deltas = (None, None, None)
        for i, (pos1, pos2) in enumerate(zip(*move_components)):
            resource = pos1.parent_resource 
            if resource is not pos2.parent_resource:
                return False
            dx, dy, constraints = resource.alignment_delta(pos1, pos2)
            if not all(c in constraints for c in self.alignment_constraints):
                return False
            alignment_deltas[i] = (dx, dy)
        ad_source, ad_tips, ad_dests = alignment_deltas
        return ad_source == ad_tips == ad_dests


class Independent8Channel(ChannelHead):

    def __init__(self):
        self.alignment_constraints = [DeckResource.align.VERTICAL]


class Standard96Channel(ChannelHead):

    def __init__(self):
        self.alignment_constraints = [DeckResource.align.STD_96]


class HamiltonDevice:
    # Just a structure with some guaranteed pieces

    def __init__(self, interface, heads):
        if not (isinstance(interface, HamiltonInterface) and all((isinstance(h, ChannelHead) for h in heads))
               and isinstance(resource_mgr, LayoutManager)):
            raise TypeError("HamiltonDevice instantiated with wrong types") # todo maybe more informative
        interface.bind_device(self)
        layout_mgr.bind_device(self)
        self.heads = heads
        for head in heads:
            head.bind_device(self)


class HamiltonAction:
    # Defined as a robot action that can be accomplished by some finite sequence of sent OEM commands.
    # HamiltonActions are nestable, and each implements execute() which might call child execute()s.
    # HamiltonActions are either instantiated with a particular HamiltonDevice or are implicitly tied to one
    # execute() is responsible for doing its own error handling.
    # Base-level actions like aspirate and dispense interpret the responses and errors from the robot device
    # A possible() method is implemented to facilitate search over HamiltonActions.

    def possible(self):
        return True

    def execute(self):
        raise NotImplementedError()


class GroupableAction(HamiltonAction):
    pass


class TipPickup(GroupableAction):

    def __init__(self, tip):
        self.tip = tip

    def execute(self):
        id = hammy.send_command(PICKUP, {'labwarePositions':str(tip.parent_resource.layout_name()) + ', ' + str(tip.index) + ';'})
        response = hammy.wait_on_response(id)


class Transfer(GroupableAction):

    def __init__(self, tip, ml, source, dest):
        self.tip = tip
        self.ml = ml
        self.source = source
        self.dest = dest
        self.action_type = HamiltonAction.MOVE

    def execute(self):
        self.pickup.execute()
        self.aspirate.execute()
        self.dispense.execute()

    def __iter__(self):
        return iter((self.tip, self.ml, self.source, self.dest))


class FlyTransfer(Transfer):

    def __init__(self, tip, ml, source, dest):
        self.tip = tip
        self.ml = ml
        self.source = source
        self.dest = dest
        self.action_type = HamiltonAction.MOVE


class Mix(HamiltonAction):

    def __init__(self, tip, ml, target, times):
        self.params = (tip, ml, target, times)
        self.action_type = HamiltonAction.MIX


class HamiltonCoordinator:

    def __init__(self, hamilton_interface, channel_heads):
        if not (isinstance(hamilton_interface, HamiltonInterface) and hamilton_interface.is_open()):
            raise ValueError('Coodinator can only start with an open HamiltonInterface')
        self.queued_actions = []
        self.heads = channel_heads
        hamilton_interface.send_command(None, INITIALIZE, block=True)
        print('Initialized Hamilton')

    def stage(self, action):
        if not isinstance(action, HamiltonAction):
            raise TypeError('Coordinator can only stage or execute HamiltonActions')
        self.queued_actions.append

    def execute(self, specific_actions=None):
        if not self.hamilton_interface.is_open():
            raise ValueError('Coodinator can only execute commands with an open HamiltonInterface')
        if specific_actions is None:
            work_list = self.queued_actions[:]
        else:
            try:
                work_list = iter(specific_actions)
            except TypeError:
                work_list = [specific_actions]
        if not work_list:
            return
        while work_list:
            action = work_list.pop(0)
            if not isinstance(action, HamiltonAction):
                raise TypeError('Coordinator can only stage or execute HamiltonActions')
            if not isinstance(action, GroupableAction):
                action.execute()
                continue
            to_move = tuple([] for h in self.heads)
            for hi, head in enumerate(self.heads):
                for move in work_list:
                    if move.source is first_move.source or move.dest is first_move.dest:
                        continue
                    if head.can_move_simul(first_move, move):
                        to_move[hi].append(move)

    def wait_for_all():
        self.hamilton_interface.block_until_clear()





# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 12:01:01 2022

@author: stefa
"""

import sys, os, time, logging, importlib
from threading import Thread

from .interface import HamiltonInterface
from .deckresource import LayoutManager, ResourceType, Plate24, Plate96, Tip96
from .oemerr import PositionError
from .interface import (INITIALIZE, PICKUP, EJECT, ASPIRATE, DISPENSE, ISWAP_GET, ISWAP_PLACE, HEPA,
WASH96_EMPTY, PICKUP96, EJECT96, ASPIRATE96, DISPENSE96, ISWAP_MOVE, MOVE_SEQ)


def resource_list_with_prefix(layout_manager, prefix, res_class, num_ress, order_key=None, reverse=False):
    def name_from_line(line):
        field = LayoutManager.layline_objid(line)
        if field:
            return field
        return LayoutManager.layline_first_field(line)
    layline_test = lambda line: LayoutManager.field_starts_with(name_from_line(line), prefix)
    res_type = ResourceType(res_class, layline_test, name_from_line)
    res_list = [layout_manager.assign_unused_resource(res_type, order_key=order_key, reverse=reverse) for _ in range(num_ress)]
    return res_list

def layout_item(lmgr, item_class, item_name):
    return lmgr.assign_unused_resource(ResourceType(item_class, item_name))

def labware_pos_str(labware, idx):
    return labware.layout_name() + ', ' + labware.position_id(idx)

def compound_pos_str(pos_tuples):
    present_pos_tups = [pt for pt in pos_tuples if pt is not None]
    return ';'.join((labware_pos_str(labware, idx) for labware, idx in present_pos_tups))

def compound_pos_str_96(labware96):
    return ';'.join((labware_pos_str(labware96, idx) for idx in range(96)))

def initialize(ham, asynch=False):
    logging.info('initialize: ' + ('a' if asynch else '') + 'synchronously initialize the robot')
    cmd = ham.send_command(INITIALIZE)
    if not asynch:
        ham.wait_on_response(cmd, raise_first_exception=True, timeout=300)
    return cmd

def hepa_on(ham, speed=15, asynch=False, **more_options):
    logging.info('hepa_on: turn on HEPA filter at ' + str(speed) + '% capacity' +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    cmd = ham.send_command(HEPA, fanSpeed=speed, **more_options)
    if not asynch:
        ham.wait_on_response(cmd, raise_first_exception=True)
    return 

def wash_empty_refill(ham, asynch=False, **more_options):
    logging.info('wash_empty_refill: empty the washer' +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    cmd = ham.send_command(WASH96_EMPTY, **more_options)
    if not asynch:
        ham.wait_on_response(cmd, raise_first_exception=True)
    return cmd

def move_plate(ham, source_plate, target_plate, gripHeight = 6, try_inversions=None):
    logging.info('move_plate: Moving plate ' + source_plate.layout_name() + ' to ' + target_plate.layout_name())
    src_pos = labware_pos_str(source_plate, 0)
    trgt_pos = labware_pos_str(target_plate, 0)
    if try_inversions is None:
        try_inversions = (0, 1)
    for inv in try_inversions:
        cid = ham.send_command(ISWAP_GET, plateLabwarePositions=src_pos, gripHeight=gripHeight, widthBefore=132, inverseGrip=inv)
        try:
            ham.wait_on_response(cid, raise_first_exception=True, timeout=120)
            break
        except PositionError:
            pass
    else:
        raise IOError
    cid = ham.send_command(ISWAP_PLACE, plateLabwarePositions=trgt_pos)
    try:
        ham.wait_on_response(cid, raise_first_exception=True, timeout=120)
    except PositionError:
        raise IOError

def move_by_seq(ham, source_plate_seq, target_plate_seq, grip_height = 0, try_inversions=None, gripForce = 2):
    logging.info('move_lid_by_seq: Moving plate ' + source_plate_seq + ' to ' + target_plate_seq)
    #src_pos = labware_pos_str(source_plate, 0)
    #trgt_pos = labware_pos_str(target_plate, 0)
    if try_inversions is None:
        try_inversions = (0, 1)
    for inv in try_inversions:
        cid = ham.send_command(ISWAP_GET, plateSequence=source_plate_seq, gripHeight=grip_height, gripForce=gripForce, inverseGrip=inv,transportMode=0, widthBefore = 132)
        try:
            ham.wait_on_response(cid, raise_first_exception=True, timeout=120)
            break
        except PositionError:
            pass
    else:
        raise IOError
    cid = ham.send_command(ISWAP_PLACE, plateSequence=target_plate_seq)
    try:
        ham.wait_on_response(cid, raise_first_exception=True, timeout=120)
    except PositionError:
        raise IOError

def channel_var(pos_tuples):
    ch_var = ['0']*16
    for i, pos_tup in enumerate(pos_tuples):
        if pos_tup is not None:
            ch_var[i] = '1'
    return ''.join(ch_var)

def tip_pick_up(ham_int, pos_tuples, **more_options):
    logging.info('tip_pick_up: Pick up tips at ' + '; '.join((labware_pos_str(*pt) if pt else '(skip)' for pt in pos_tuples)) +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    num_channels = len(pos_tuples)
    if num_channels > 8:
        raise ValueError('Can only pick up 8 tips at a time')
    ch_patt = channel_var(pos_tuples)
    labware_poss = compound_pos_str(pos_tuples)
    ham_int.wait_on_response(ham_int.send_command(PICKUP,
        labwarePositions=labware_poss,
        channelVariable=ch_patt,
        **more_options), raise_first_exception=True)

def tip_eject(ham_int, pos_tuples=None, **more_options):
    if pos_tuples is None:
        logging.info('tip_eject: Eject tips to default waste' + ('' if not more_options else ' with extra options ' + str(more_options)))
        more_options['useDefaultWaste'] = 1
        dummy = Tip96('')
        pos_tuples = [(dummy, 0)] * 8
    else:
        logging.info('tip_eject: Eject tips to ' + '; '.join((labware_pos_str(*pt) if pt else '(skip)' for pt in pos_tuples)) +
                ('' if not more_options else ' with extra options ' + str(more_options)))
    num_channels = len(pos_tuples)
    if num_channels > 8:
        raise ValueError('Can only eject up to 8 tips')
    ch_patt = channel_var(pos_tuples)
    labware_poss = compound_pos_str(pos_tuples)
    ham_int.wait_on_response(ham_int.send_command(EJECT,
        labwarePositions=labware_poss,
        channelVariable=ch_patt,
        **more_options), raise_first_exception=True)

def assert_parallel_nones(list1, list2):
    if not (len(list1) == len(list2) and all([(i1 is None) == (i2 is None) for i1, i2 in zip(list1, list2)])):
        raise ValueError('Lists must have parallel None entries')

default_liq_class = 'HighVolumeFilter_Water_DispenseJet_Empty_with_transport_vol'


def aspirate(ham_int, pos_tuples, vols, **more_options):
    assert_parallel_nones(pos_tuples, vols)
    logging.info('aspirate: Aspirate volumes ' + str(vols) + ' from positions [' +
            '; '.join((labware_pos_str(*pt) if pt else '(skip)' for pt in pos_tuples)) +
            (']' if not more_options else '] with extra options ' + str(more_options)))
    if len(pos_tuples) > 8:
        raise ValueError('Can only aspirate with 8 channels at a time')
    if 'liquidClass' not in more_options:
        more_options.update({'liquidClass':default_liq_class})
    ham_int.wait_on_response(ham_int.send_command(ASPIRATE,
        channelVariable=channel_var(pos_tuples),
        labwarePositions=compound_pos_str(pos_tuples),
        volumes=[v for v in vols if v is not None],
        **more_options), raise_first_exception=True)

def dispense(ham_int, pos_tuples, vols, **more_options):
    assert_parallel_nones(pos_tuples, vols)
    logging.info('dispense: Dispense volumes ' + str(vols) + ' into positions [' +
            '; '.join((labware_pos_str(*pt) if pt else '(skip)' for pt in pos_tuples)) +
            (']' if not more_options else '] with extra options ' + str(more_options)))
    if len(pos_tuples) > 8:
        raise ValueError('Can only aspirate with 8 channels at a time')
    if 'liquidClass' not in more_options:
        more_options.update({'liquidClass':default_liq_class})
    ham_int.wait_on_response(ham_int.send_command(DISPENSE,
        channelVariable=channel_var(pos_tuples),
        labwarePositions=compound_pos_str(pos_tuples),
        volumes=[v for v in vols if v is not None],
        **more_options), raise_first_exception=True)

def tip_pick_up_96(ham_int, tip96, **more_options):
    logging.info('tip_pick_up_96: Pick up tips at ' + tip96.layout_name() +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    labware_poss = compound_pos_str_96(tip96)
    ham_int.wait_on_response(ham_int.send_command(PICKUP96,
        labwarePositions=labware_poss,
        **more_options), raise_first_exception=True)

def tip_eject_96(ham_int, tip96=None, **more_options):
    logging.info('tip_eject_96: Eject tips to ' + (tip96.layout_name() if tip96 else 'default waste') +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    if tip96 is None:
        labware_poss = ''
        more_options.update({'tipEjectToKnownPosition':2}) # 2 is default waste
    else:   
        labware_poss = compound_pos_str_96(tip96)
    ham_int.wait_on_response(ham_int.send_command(EJECT96,
        labwarePositions=labware_poss,
        **more_options), raise_first_exception=True)

def aspirate_96(ham_int, plate96, vol, **more_options):
    logging.info('aspirate_96: Aspirate volume ' + str(vol) + ' from ' + plate96.layout_name() +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    if 'liquidClass' not in more_options:
        more_options.update({'liquidClass':default_liq_class})
    ham_int.wait_on_response(ham_int.send_command(ASPIRATE96,
        labwarePositions=compound_pos_str_96(plate96),
        aspirateVolume=vol,
        **more_options), raise_first_exception=True)

def dispense_96(ham_int, plate96, vol, **more_options):
    logging.info('dispense_96: Dispense volume ' + str(vol) + ' into ' + plate96.layout_name() +
            ('' if not more_options else ' with extra options ' + str(more_options)))
    if 'liquidClass' not in more_options:
        more_options.update({'liquidClass':default_liq_class})
    ham_int.wait_on_response(ham_int.send_command(DISPENSE96,
        labwarePositions=compound_pos_str_96(plate96),
        dispenseVolume=vol,
        **more_options), raise_first_exception=True)


def move_sequence(ham_int, sequence, xDisplacement=0, yDisplacement=0, zDisplacement=0):
    cid = ham_int.send_command(MOVE_SEQ, inputSequence=sequence, xDisplacement=xDisplacement, yDisplacement=yDisplacement, zDisplacement=zDisplacement)
    ham_int.wait_on_response(cid, raise_first_exception=True, timeout=120)

class StderrLogger:
    def __init__(self, level):
        self.level = level
        self.stderr = sys.stderr

    def write(self, message):
        self.stderr.write(message)
        if message.strip():
            self.level(message.replace('\n', ''))

def add_stderr_logging(logger_name=None):
    logger = logging.getLogger(logger_name) # root logger if None
    sys.stderr = StderrLogger(logger.error)
    
def normal_logging(ham_int, method_local_dir):
    local_log_dir = os.path.join(method_local_dir, 'log')
    if not os.path.exists(local_log_dir):
        os.mkdir(local_log_dir)
    main_logfile = os.path.join(local_log_dir, 'main.log')
    logging.basicConfig(filename=main_logfile, level=logging.DEBUG, format='[%(asctime)s] %(name)s %(levelname)s %(message)s')
    #add_robot_level_log()
    add_stderr_logging()
    import __main__
    for banner_line in log_banner('Begin execution of ' + __main__.__file__):
        logging.info(banner_line)
    ham_int.set_log_dir(os.path.join(local_log_dir, 'hamilton.log'))


def run_async(funcs):
    def go():
        try:
            iter(funcs)
        except TypeError:
            funcs()
            return
        for func in funcs:
            func()
    func_thread = Thread(target=go, daemon=True)
    func_thread.start()
    return func_thread

def run_async_dict(func):
    logging.info("running async line 427 pace_util_stefan_dev.py")
    def go():
        func['function'](func['arguments'])
        return
    func_thread = Thread(target=go, daemon=True)
    func_thread.start()
    return func_thread


def yield_in_chunks(sliceable, n):
    sliceable = list(sliceable)
    start_pos = 0
    end_pos = n
    while start_pos < len(sliceable):
        yield sliceable[start_pos:end_pos]
        start_pos, end_pos = end_pos, end_pos + n

def log_banner(banner_text):
    l = len(banner_text)
    margin = 5
    width = l + 2*margin + 2
    return ['#'*width,
            '#' + ' '*(width - 2) + '#',
            '#' + ' '*margin + banner_text + ' '*margin + '#',
            '#' + ' '*(width - 2) + '#',
            '#'*width]



import os
from pyhamilton import PACKAGE_PATH

LAY_BACKUP_DIR = os.path.join(PACKAGE_PATH, 'LAY-BACKUP')
if not os.path.exists(LAY_BACKUP_DIR):
    os.mkdir(LAY_BACKUP_DIR)
OEM_STAR_PATH = os.path.join(PACKAGE_PATH, 'star-oem')
if not (os.path.exists(OEM_STAR_PATH)
		and os.path.exists(os.path.join(OEM_STAR_PATH, 'RunHSLExecutor.dll'))
		and os.path.exists(os.path.join(OEM_STAR_PATH, 'HSLHttp.dll'))):
    raise FileNotFoundError('pyhamilton requires .../site-packages/pyhamilton/STAR-OEM, distributed separately.')
OEM_LAY_PATH = os.path.join(OEM_STAR_PATH, 'VENUS_Method', 'STAR_OEM_Test.lay')
OEM_HSL_PATH = os.path.join(OEM_STAR_PATH, 'VENUS_Method', 'STAR_OEM_noFan.hsl')
OEM_RUN_EXE_PATH = 'C:\\Program Files (x86)\\HAMILTON\\Bin\\HxRun.exe'



import time, json, signal, os, requests, string, logging, subprocess, win32gui, win32con
from http import server
from threading import Thread
from multiprocessing import Process
from pyhamilton import OEM_RUN_EXE_PATH, OEM_HSL_PATH
from .oemerr import * #TODO: specify
from .defaultcmds import defaults_by_cmd

class HamiltonCmdTemplate:
    """
    Formatter object to create valid `pyhamilton` command dicts.

    Use of this class to assemble JSON pyhamilton commands enables keyword access to command attributes, which cuts down on string literals. It also helps to fail malformed commands early, before they are sent.

    Several default `HamiltonCmdTemplate`s are defined in `pyhamilton.defaultcmds`, such as `INITIALIZE`, `ASPIRATE`, and `DISPENSE`. Casual users will most likely never need to manually instantiate a HamiltonCmdTemplate.
    """

    @staticmethod
    def unique_id():
        """Return a "uniqe" hexadecimal string (`'0x...'`) based on time of call."""
        return hex(int((time.time()%3600e4)*1e6))

    def __init__(self, cmd_name, params_list):
        """
        Creates a `HamiltonCmdTemplate` with a command name and required parameters.

        The command name must be one of the command names accepted by the
        `pyhamilton` interpreter and a list of expected parameters for this command.

        Args:
          cmd_name (str): One of the set of string literals recognized as command names
            by the `pyhamilton` interpreter, e.g. `'mph96Dispense'`. See `pyhamilton.defaultcmds` for examples.
          params_list (list): exact list of string parameters that must have associated
            values for the command to be valid, other than those that are always present
            (`'command'` and `'id'`)
        """
        self.cmd_name = cmd_name
        self.params_list = params_list
        if cmd_name in defaults_by_cmd:
            const_name, default_dict = defaults_by_cmd[cmd_name]
            self.defaults = {k:v for k, v in default_dict.items() if v is not None}
        else:
            self.defaults = {}

    def assemble_cmd(self, *args, **kwargs):
        """
        Use keyword args to assemble this command. Default values auto-filled.

        Args:
          kwargs (dict): map of any parameters (str) to values that should be different
            from the defaults supplied for this command in `pyhamilton.defaultcmds`
        """
        if args:
            raise ValueError('assemble_cmd can only take keyword arguments.')
        assembled_cmd = {'command':self.cmd_name, 'id':HamiltonCmdTemplate.unique_id()}
        assembled_cmd.update(self.defaults)
        assembled_cmd.update(kwargs)
        self.assert_valid_cmd(assembled_cmd)
        return assembled_cmd

    def assert_valid_cmd(self, cmd_dict):
        """Validate a finished command. Do nothing if it is valid.

        `ValueError` will be raised if the supplied command did not have all required
        parameters for this command, as well as values for keys `'id'` and `'command'`, which
        are always required.

        Args:
          cmd_dict (dict): A fully assembled `pyhamilton` command

        Raises:
          ValueError: The command dict is not ready to send. Specifics of mismatch
            summarized in exception description.
        """
        prefix = 'Assert valid command "' + self.cmd_name + '" failed: '
        if 'id' not in cmd_dict:
            raise ValueError(prefix + 'no key "id"')
        if 'command' not in cmd_dict:
            raise ValueError(prefix + 'no key "command"')
        if cmd_dict['command'] != self.cmd_name:
            raise ValueError(prefix + 'command name "' + cmd_dict['command'] + '" does not match')
        needs = set(['command', 'id'])
        needs.update(self.params_list)
        givens = set(cmd_dict.keys())
        if givens != needs:
            prints = [prefix + 'template parameter keys (left) do not match given keys (right)\n']
            q_mark = ' (?)  '
            l_col_space = 4
            r_col_space = max((len(key) for key in needs)) + len(q_mark) + 1
            needs_l = sorted(list(needs))
            givens_l = sorted(list(givens))
            while needs_l or givens_l:
                if needs_l:
                    lval = needs_l.pop(0)
                    if lval not in givens:
                        lval = q_mark + lval
                else:
                    lval = ''
                if givens_l:
                    rval = givens_l.pop(0)
                    if rval not in needs:
                        rval = q_mark + rval
                else:
                    rval = ''
                prints.append(' '*l_col_space + lval + ' '*(r_col_space - len(lval)) + rval)
            raise ValueError('\n'.join(prints))

_builtin_templates_by_cmd = {}

for cmd in defaults_by_cmd:
    const_name, default_dict = defaults_by_cmd[cmd]
    const_template = HamiltonCmdTemplate(cmd, list(default_dict.keys()))
    globals()[const_name] = const_template
    _builtin_templates_by_cmd[cmd] = const_template

def _make_new_hamilton_serv_handler(resp_indexing_fn):
    """Make HTTP request handler to aggregate responses according to an index function.

    A new class is defined each time, bound to a specific indexing function, to keep it
    agnostic to any particular indexing scheme. In practice, the current implementation
    uses the value of the key 'id'; that is the scheme for the `pyhamilton` interpreter.

    Attributes:
      indexed_responses (dict): aggregated responses received by this handler, keyed by
        the values returned by `resp_indexing_fn`.

    Args:
      resp_indexing_fn (Callable[[str], Hashable]): Called on every response body (str)
        to extract a hashable index. Later, the response can be retrieved by this index
        from `indexed_responses`.

    """

    class HamiltonServerHandler(server.BaseHTTPRequestHandler):
        _send_queue = []
        indexed_responses = {}
        indexing_fn = resp_indexing_fn
        MAX_QUEUED_RESPONSES = 1000

        @staticmethod
        def send_str(cmd_str):
            if not isinstance(cmd_str, b''.__class__):
                if isinstance(cmd_str, ''.__class__):
                    cmd_str = cmd_str.encode()
                else:
                    raise ValueError('send_command can only send strings, not ' + str(cmd_str))
            HamiltonServerHandler._send_queue.append(cmd_str)

        @staticmethod
        def has_queued_cmds():
            return bool(HamiltonServerHandler._send_queue)

        @staticmethod
        def pop_response(idx):
            ir = HamiltonServerHandler.indexed_responses
            if idx not in ir:
                raise KeyError('No response received with index ' + str(idx))
            return ir.pop(idx).decode()

        def _set_headers(self):
            self.send_response(200)
            self.send_header('Content-type', 'text/HTML')
            self.end_headers()

        def do_GET(self):
            sq = HamiltonServerHandler._send_queue
            response_to_send = sq.pop(0) if sq else b''
            self._set_headers()
            self.wfile.write(response_to_send)

        def do_HEAD(self):
            self._set_headers()

        def do_POST(self):
            content_len = int(self.headers.get('content-length', 0))
            post_body = self.rfile.read(content_len)
            self._set_headers()
            self.wfile.write(b'<html><body><h1>POST!</h1></body></html>')
            ir = HamiltonServerHandler.indexed_responses
            index = HamiltonServerHandler.indexing_fn(post_body)
            if index is None:
                return
            ir[index] = post_body

        def log_message(self, *args, **kwargs):
            pass

    return HamiltonServerHandler

def run_hamilton_process():
    """Start the interpreter in a separate python process.

    Starts the pyhamilton interpreter, which is an HSL file to be passed to the
    RunHSLExecutor.exe executable from Hamilton. This should always be done in a
    separate python process using the subprocess module, not a Thread.
    """
    import clr
    from pyhamilton import OEM_STAR_PATH, OEM_HSL_PATH
    clr.AddReference(os.path.join(OEM_STAR_PATH, 'RunHSLExecutor'))
    clr.AddReference(os.path.join(OEM_STAR_PATH, 'HSLHttp'))
    try:
        from RunHSLExecutor import Class1
    except ModuleNotFoundError:
        raise RuntimeError('RunHSLExecutor DLLs successfully located, but an internal '
                           'error prevented import as a CLR module. You might be '
                           'missing the standard Hamilton software suite HSL '
                           'executables, their DLLs may not be registered with Windows, '
                           'or they may not be located in the expected system '
                           'directory.')
    C = Class1()
    C.StartMethod(OEM_HSL_PATH)
    try:
        while True:
            pass # Send external signal to end process
    except:
        pass

_block_numfield = 'Num'
_block_mainerrfield = 'MainErr'
BLOCK_FIELDS = _block_numfield, _block_mainerrfield, 'SlaveErr', 'RecoveryBtnId', 'StepData', 'LabwareName', 'LabwarePos'
_block_field_types = int, int, int, int, str, str, str

class HamiltonInterface:
    """Main class to automatically set up and tear down an interface to a Hamilton robot.

    HamiltonInterface is the primary class offered by this module. It creates a Hamilton
    HSL background process running the `pyhamilton` interpreter, along with a `localhost`
    connection to act as a bridge. It is recommended to create a `HamiltonInterface` using
    a `with:` block to ensure proper startup and shutdown of its async components, even if
    exceptions are raised. It may be used with explicit `start()` and `stop()` calls.

      Typical usage:

      ```
      with HamiltonInterface() as ham_int:
          cmd_id = ham_int.send_command(INITIALIZE)
          ...
          ham_int.wait_on_response(cmd_id)
          ...
      ```
    """

    known_templates = _builtin_templates_by_cmd
    default_port = 3221
    default_address = '127.0.0.1' # localhost

    class HamiltonServerThread(Thread):
        """Private threaded local HTTP server with graceful shutdown flag."""

        def __init__(self, address, port):
            Thread.__init__(self)
            self.server_address = (address, port)
            self.should_continue = True
            self.exited = False
            def index_on_resp_id(response_str):
                try:
                    response = json.loads(response_str)
                    if 'id' in response:
                        return response['id']
                    return None
                except json.decoder.JSONDecodeError:
                    return None
            self.server_handler_class = _make_new_hamilton_serv_handler(index_on_resp_id)
            self.httpd = None

        def run(self):
            self.exited = False
            self.httpd = server.HTTPServer(self.server_address, self.server_handler_class)
            while self.should_continue:
                self.httpd.handle_request()
            self.exited = True

        def disconnect(self):
            self.should_continue = False

        def has_exited(self):
            return self.exited

    def __init__(self, address=None, port=None, simulate=False):
        self.address = HamiltonInterface.default_address if address is None else address
        self.port = HamiltonInterface.default_port if port is None else port
        self.simulate = simulate
        self.server_thread = None
        self.oem_process = None
        self.active = False
        self.logger = None
        self.log_queue = []

    def start(self):
        """Starts the extra processes, threads, and servers for the Hamilton connection.

        Launches: 1) the pyhamilton interpreter using the Hamilton Run Control
        executable, either in the background for normal use, or in the foreground with a
        GUI for simulation; 2) a local HTTP server to ferry messages between the python
        module and the interpreter.

        When used with a `with:` block, called automatically upon entering the block.
        """

        if self.active:
            return
        self.log('starting a Hamilton interface')
        if self.simulate:
            sim_window_handle = None
            try:
                sim_window_handle = win32gui.FindWindow(None, 'Hamilton Run Control - ' + os.path.basename(OEM_HSL_PATH))
            except win32gui.error:
                pass
            if sim_window_handle:
                try:
                    win32gui.SendMessage(sim_window_handle, win32con.WM_CLOSE, 0, 0)
                    os.system('taskkill /f /im HxRun.exe')
                except win32gui.error:
                    self.stop()
                    self.log_and_raise(OSError('Simulator already open'))
            subprocess.Popen([OEM_RUN_EXE_PATH, OEM_HSL_PATH])
            self.log('started the oem application for simulation')
        else:
            self.oem_process = Process(target=run_hamilton_process, args=())
            self.oem_process.start()
            self.log('started the oem process')
        self.server_thread = HamiltonInterface.HamiltonServerThread(self.address, self.port)
        self.server_thread.start()
        self.log('started the server thread')
        self.active = True

    def stop(self):
        """Stop this HamiltonInterface and clean up associated async processes.

        Kills the pyhamilton interpreter subprocess and executable and stops the local
        web server thread.

        When used with a `with` block, called automatically on exiting the block.
        """

        if not self.active:
            return
        try:
            if self.simulate:
                self.log('sending end run command to simulator')
                try:
                    self.wait_on_response(self.send_command(command='end', id=hex(0)), timeout=1.5)
                except HamiltonTimeoutError:
                    pass
            else:
                for i in range(2):
                    try:
                        os.kill(self.oem_process.pid, signal.SIGTERM)
                        self.log('sent sigterm to oem process')
                        self.oem_process.join()
                        self.log('oem process exited')
                        break
                    except PermissionError:
                        self.log('permission denied, trying again...', 'warn')
                        time.sleep(2)
                else:
                    self.log('Could not kill oem process, moving on with shutdown', 'warn')
        finally:
            self.active = False
            self.server_thread.disconnect()
            self.log('disconnected from server')
            time.sleep(.1)
            if not self.server_thread.has_exited():
                self.log('server did not exit yet, sending dummy request to exit its loop')
                session = requests.Session()
                adapter = requests.adapters.HTTPAdapter(max_retries=20)
                session.mount('http://', adapter)
                session.get('http://' + HamiltonInterface.default_address + ':' + str(HamiltonInterface.default_port))
                self.log('dummy get request sent to server')
            self.server_thread.join()
            self.log('server thread exited')

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, tb):
        self.stop()

    def is_open(self):
        """Return `True` if the HamiltonInterface has been started and not stopped."""
        return self.active

    def send_command(self, template=None, block_until_sent=False, *args, **cmd_dict): # returns unique id of command
        """Add a command templated after HamiltonCmdTemplate to the server send queue.

        Args:
          template (HamiltonCmdTemplate): Optional; a template to provide default
            arguments not specified in `cmd_dict`. 
          block_until_sent (bool): Optional; if `True`, wait for all queued messages,
            including this one, to get picked up by the local server and sent across
            the HTTP connection, before returning. Default is False.
          cmd_dict (dict): keyword arguments to be forwarded to `template` when building
            the command, overriding its defaults. If `template` not given, cmd_dict must
            either have a 'command' key with value matching one of the command names in
            `defaultcmds` and might be missing an 'id' key, or itself be a fully formed
            and correct pyhamilton command with its own 'id' key.

        Returns:
          unique id (str) of the command that can be used to index it later, either
            newly generated or same as originally present in cmd_dict.
        """
        if not self.is_open():
            self.log_and_raise(RuntimeError('Cannot send a command from a closed HamiltonInterface'))
        if template is None:
            if 'command' not in cmd_dict:
                self.log_and_raise(ValueError('Command dicts from HamiltonInterface must have a \'command\' key'))
            cmd_name = cmd_dict['command']
            if cmd_name in HamiltonInterface.known_templates:
                # raises if this is a known command but some fields in cmd_dict are invalid
                send_cmd_dict = HamiltonInterface.known_templates[cmd_name].assemble_cmd(**cmd_dict)
            else:
                send_cmd_dict = cmd_dict
        else:
            send_cmd_dict = template.assemble_cmd(**cmd_dict)
        if 'id' not in send_cmd_dict:
            self.log_and_raise(ValueError("Command dicts sent from HamiltonInterface must have a unique id with key 'id'"))
        self.server_thread.server_handler_class.send_str(json.dumps(send_cmd_dict))
        if block_until_sent:
            self._block_until_sq_clear()
        return send_cmd_dict['id']

    def wait_on_response(self, id, timeout=0, raise_first_exception=False, return_data=False):
        """Wait and do not return until the response for the specified id comes back.

        When the command corresponding to `id` regards multiple distinct pipette channels
        or devices, responses may contain encoded errors that might be different for
        different channels or devices. For this reason, the default behavior of
        `wait_on_response` is to not raise exceptions, but to delegate handling
        exceptions to the caller. For convenience, this method can optionally raise the
        first exception it encounters, often a useful behavior for succinct scripted
        commands that regard only one device, when raise_first_exception is `True`.

        Args:
          id (str): The unique id of a previously sent command
          timeout (float): Optional; maximum time in seconds to wait before raising
            `HamiltonTimeoutError`. Default is no timeout (forever).
          raise_first_exception: Optional; if True, may raise if there is an error
            encoded in the response. Default is False.

        Returns:
          The response dict from the hamilton interpreter.

        Raises:
          `HamiltonTimeoutError`: after `timeout` seconds elapse with no response, if
          `timeout` was specified.
        """

        if timeout:
            start_time = time.time()
        else:
            start_time = float('inf')

        response_tup = None
        while time.time() - start_time < timeout:
            try:
                response_tup = self.pop_response(id, raise_first_exception, return_data)
            except KeyError:
                pass
            if response_tup is not None:
                return response_tup
            time.sleep(.1)
        self.log_and_raise(HamiltonTimeoutError('Timed out after ' + str(timeout) + ' sec while waiting for response id ' + str(id)))

    def pop_response(self, id, raise_first_exception=False, return_data=False):
        """Remove and return the response with the specified id from the response queue.

        If there is a response, remove it and return the Hamilton-formatted response
        dict, like that returned from `HamiltonInterface.parse_hamilton_return`. Otherwise, raise
        `KeyError`.

        Args:
          id (str): Unique id of the command that initiated the response
          raise_first_exception (bool): Optional; forwarded to `wait_on_response`.
            Default is `False`.

        Returns:
          A 2-tuple:

            1. parsed response block dict from Hamilton as in `parse_hamilton_return`
            2. Error map, a dict mapping int keys (data block Num field) that had
                exceptions, if any, to an exception that was coded in block;
                `None` to any error not associated with a block. `{}` if no error

        Raises:
          KeyError: if `id` has no matching response in the queue.
        """

        try:
            response = self.server_thread.server_handler_class.pop_response(id)
        except KeyError:
            raise KeyError('No Hamilton interface response indexed for id ' + str(id))
        if return_data:
            print(type(response))
            print(response)
            response = json.loads(response)
            response_data=response['step-return1']
            return response_data
        errflag, blocks = self.parse_hamilton_return(response)
        err_map = {}
        if errflag:
            for blocknum in sorted(blocks.keys()):
                errcode = blocks[blocknum][_block_mainerrfield]
                if errcode != 0:
                    self.log('Exception encoded in Hamilton return.', 'warn')
                    try:
                        decoded_exception = HAMILTON_ERROR_MAP[errcode]()
                    except KeyError:
                        self.log_and_raise(InvalidErrCodeError('Response returned had an unknown error code: ' + str(errcode)))
                    self.log('Exception: ' + repr(decoded_exception), 'warn')
                    if raise_first_exception:
                        self.log('Raising first exception.', 'warn')
                        raise decoded_exception
                    err_map[blocknum] = decoded_exception
            else:
                unknown_exc = HamiltonStepError('Hamilton step did not execute correctly; no error code given.')
                err_map[None] = unknown_exc
                if raise_first_exception:
                    self.log('Raising first exception; exception has no error code.', 'warn')
                    raise unknown_exc
        return blocks, err_map

    def _block_until_sq_clear(self):
        while HamiltonServerHandler.has_queued_cmds():
            pass

    def parse_hamilton_return(self, return_str):
        """
        Return a 2-tuple:

        - [0] errflag: any error code present in response

        - [1] Block map: dict mapping int keys to dicts with str keys (MainErr, SlaveErr, RecoveryBtnId, StepData, LabwareName, LabwarePos)

        Result value 3 is the field that is returned by the OEM interface.

        "Result value 3 contains one error flag (ErrFlag) and the block data package."

        ### Data Block Format Rules

        - The error flag is set once only at the beginning of result value 3. The error flag
        does not belong to the block data but may be used for a simpler error recovery.
        If this flag is set, an error code has been set in any of the block data entries.

        - Each block data package starts with the opening square bracket character '['

        - The information within the block data package is separated by the comma delimiter ','

        - Block data information may be empty; anyway a comma delimiter is set.

        - The result value may contain more than one block data package.

        - Block data packages are returned independent of Num value ( unsorted ).

        ### Block data information

        - Num 
            - Step depended information (e.g. the channel number, a loading position etc.).

            - Note: The meaning and data type for this information is described in the corresponding help of single step.

        - MainErr
            - Main error code which occurred on instrument.

        - SlaveErr
            - Detailed error code of depended slave (e.g. auto load, washer etc.).

        - RecoveryBtnId
            - Recovery which has been used to handle this error.

        - StepData
            - Step depended information, e.g. the barcode read, the volume aspirated etc.

            - Note: The meaning and data type for this information is described in the corresponding help of single step.

        - LabwareName
            - Labware name of used labware.

        - LabwarePos
            - Used labware position.
        """

        def raise_parse_error():
            msg = 'Could not parse response ' + repr(return_str)
            self.log(msg, 'error')
            raise HamiltonReturnParseError(msg)

        try:
            block_data_str = str(json.loads(return_str)['step-return1'])
        except KeyError:
            raise_parse_error()
        blocks = block_data_str.split('[')
        try:
            errflag = int(blocks.pop(0)) != 0
        except ValueError:
            raise_parse_error()

        blocks_by_blocknum = {}
        any_error_code = False
        for block_str in blocks:
            field_vals = block_str.split(',')
            if len(field_vals) != len(BLOCK_FIELDS):
                raise_parse_error()
            try:
                block_contents = {field:cast(val) for field, cast, val in zip(BLOCK_FIELDS, _block_field_types, field_vals)}
            except ValueError:
                raise_parse_error()
            if block_contents[_block_mainerrfield] != 0:
                any_error_code = True
            blocks_by_blocknum[block_contents.pop(_block_numfield)] = block_contents
        if blocks and errflag != any_error_code:
            raise_parse_error()

        return errflag, blocks_by_blocknum

    def set_log_dir(self, log_dir):
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        hdlr = logging.FileHandler(log_dir)
        formatter = logging.Formatter('[%(asctime)s] %(name)s %(levelname)s %(message)s')
        hdlr.setFormatter(formatter)
        self.logger.addHandler(hdlr)
        self._dump_log_queue()

    def log(self, msg, msg_type='info'):
        self.log_queue.append((msg, msg_type))
        self._dump_log_queue()

    def _dump_log_queue(self):
        if self.logger is None:
            return
        log_actions = {'error':self.logger.error,
                      'warn':self.logger.warn,
                      'debug':self.logger.debug,
                      'info':self.logger.info,
                      'critical':self.logger.critical}
        while self.log_queue:
            msg, msg_type = self.log_queue.pop(0)
            log_actions.get(msg_type.lower(), self.logger.info)(msg) # prints if no log path set

    def log_and_raise(self, err):
        self.log(repr(err), 'error')
        raise err



from setuptools import setup, find_packages

try:
    print("here")
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
    print(long_description)
except(IOError, ImportError):
    long_description = open('README.md').read()

setup(
    name='pyhamilton',
    version='1.483',
    packages=find_packages(exclude=['tests*', 'examples*']),
    license='MIT',
    description='Python for Hamilton liquid handling robots',
    long_description='Forthcoming due to markdown incompatibility',
    install_requires=['requests', 'pythonnet', 'pywin32', 'pyserial'],
    package_data={'pyhamilton': ['star-oem/*', 'star-oem/VENUS_Method/*', 'bin/*','library/*',
                                 'library/HSLInhecoTEC/*','library/HSLAppsLib/*','library/ASWStandard/*',
                                 'library/ASWStandard/ASWGlobal/*','library/DaisyChainedTiltModule/*','library/SchedulingDev/*',]},
    url='https://github.com/dgretton/pyhamilton.git',
    author='Dana Gretton',
    author_email='dgretton@mit.edu',
    entry_points={
        'console_scripts': [
            'pyhamilton-new-project = pyhamilton.__init__:create_project',
            'pyhamilton-configure = pyhamilton.__init__:autoconfig'
        ],
    },
)



"""`pyhamilton`-specific exception definitions.
"""
class HamiltonError(Exception):
    """
    Exceptions raised in package pyhamilton
    """
    pass

###########################################
### BEGIN HAMILTON DECK RESOURCE ERRORS ###
###########################################

class HamiltonDeckResourceError(HamiltonError):
    """
    Error with any deck object in interface with robot.
    """
    pass

class ResourceUnavailableError(HamiltonDeckResourceError):
    """
    Layout manager found deck resource type not present or all of this type assigned
    """
    pass

#######################################
### BEGIN HAMILTON INTERFACE ERRORS ###
#######################################

class HamiltonInterfaceError(HamiltonError):
    """
    Error in any phase of communication with robot.
    """
    pass

class HamiltonTimeoutError(HamiltonInterfaceError):
    """
    An asynchronous request to the Hamilton robot timed out.
    """
    pass

class InvalidErrCodeError(HamiltonInterfaceError):
    """
    Error code returned from instrument not known.
    """
    pass

class HamiltonReturnParseError(HamiltonInterfaceError):
    """
    Return string from instrument was malformed.
    """
    pass

########################################################
### BEGIN HAMILTON CODED STEP ERRORS, CODE MAP BELOW ###
########################################################

class HamiltonStepError(HamiltonError):
    """
    Errors in steps executed by VENUS software coded in the Hamilton error specification.
    """
    pass

class HamiltonSyntaxError(HamiltonStepError):
    """
    There is a wrong set of parameters or parameter ranges.
    """
    pass
 
class HardwareError(HamiltonStepError):
    """
    Steps lost on one or more hardware components, or component not initialized or not functioning.
    """
    pass
 
class NotExecutedError(HamiltonStepError):
    """
    There was an error in previous part command.
    """
    pass

class ClotError(HamiltonStepError):
    """
    Blood clot detected.
    """
    pass

class BarcodeError(HamiltonStepError):
    """
    Barcode could not be read or is missing.
    """
    pass

class InsufficientLiquidError(HamiltonStepError):
    """
    Not enough liquid available.
    """
    pass

class TipPresentError(HamiltonStepError):
    """
    A tip has already been picked up.
    """
    pass

class NoTipError(HamiltonStepError):
    """
    Tip is missing or not picked up.
    """
    pass

class NoCarrierError(HamiltonStepError):
    """
    No carrier present for loading.
    """
    pass

class ExecutionError(HamiltonStepError):
    """
    A step or a part of a step could not be processed.
    """
    pass

class PressureLLDError(HamiltonStepError):
    """
    A dispense with pressure liquid level detection is not allowed.
    """
    pass

class CalibrateError(HamiltonStepError):
    """
    No capacitive signal detected during carrier calibration procedure.
    """
    pass

class UnloadError(HamiltonStepError):
    """
    Not possible to unload the carrier due to occupied loading tray position.
    """
    pass

class PressureLLDError(HamiltonStepError):
    """
    Pressure liquid level detection in a consecutive aspiration is not allowed.
    """
    pass

class ParameterError(HamiltonStepError):
    """
    Dispense in jet mode with pressure liquid level detection is not allowed.
    """
    pass

class CoverOpenError(HamiltonStepError):
    """
    Cover not closed or can not be locked.
    """
    pass

class ImproperAspirationOrDispenseError(HamiltonStepError):
    """
    The pressure-based aspiration / dispensation control reported an error ( not enough liquid ).
    """
    pass

class WashLiquidError(HamiltonStepError):
    """
    Waste full or no more wash liquid available.
    """
    pass

class TemperatureError(HamiltonStepError):
    """
    Incubator temperature out of range.
    """
    pass

class TADMOvershotError(HamiltonStepError):
    """
    Overshot of limits during aspirate or dispense.

    Note:

    On aspirate this error is returned as main error 17.

    On dispense this error is returned as main error 4.
    """
    pass

class LabwareError(HamiltonStepError):
    """
    Labware not available.
    """
    pass

class LabwareGrippedError(HamiltonStepError):
    """
    Labware already gripped.
    """
    pass

class LabwareLostError(HamiltonStepError):
    """
    Labware lost during transport.
    """
    pass

class IllegalTargetPlatePositionError(HamiltonStepError):
    """
    Cannot place plate, plate was gripped in a wrong direction.
    """
    pass

class IllegalInterventionError(HamiltonStepError):
    """
    Cover was opened or a carrier was removed manually.
    """
    pass

class TADMUndershotError(HamiltonStepError):
    """
    Undershot of limits during aspirate or dispense.

    Note:

    On aspirate this error is returned as main error 4.

    On dispense this error is returned as main error 17.
    """
    pass

class PositionError(HamiltonStepError):
    """
    The position is out of range.
    """
    pass

class UnexpectedcLLDError(HamiltonStepError):
    """
    The cLLD detected a liquid level above start height of liquid level search.
    """
    pass

class AreaAlreadyOccupiedError(HamiltonStepError):
    """
    Instrument region already reserved.
    """
    pass

class ImpossibleToOccupyAreaError(HamiltonStepError):
    """
    A region on the instrument cannot be reserved.
    """
    pass

class AntiDropControlError(HamiltonStepError):
    """
    Anti drop controlling out of tolerance.
    """
    pass

class DecapperError(HamiltonStepError):
    """
    Decapper lock error while screw / unscrew a cap by twister channels.
    """
    pass

class DecapperHandlingError(HamiltonStepError):
    """
    Decapper station error while lock / unlock a cap.
    """
    pass

class SlaveError(HamiltonStepError):
    """
    Slave error.
    """
    pass

class WrongCarrierError(HamiltonStepError):
    """
    Wrong carrier barcode detected.
    """
    pass

class NoCarrierBarcodeError(HamiltonStepError):
    """
    Carrier barcode could not be read or is missing.
    """
    pass

class LiquidLevelError(HamiltonStepError):
    """
    Liquid surface not detected.

    This error is created from main / slave error 06/70, 06/73 and 06/87.
    """
    pass

class NotDetectedError(HamiltonStepError):
    """
    Carrier not detected at deck end position.
    """
    pass

class NotAspiratedError(HamiltonStepError):
    """
    Dispense volume exceeds the aspirated volume.

    This error is created from main / slave error 02/54.
    """
    pass

class ImproperDispensationError(HamiltonStepError):
    """
    The dispensed volume is out of tolerance (may only occur for Nano Pipettor Dispense steps). 

    This error is created from main / slave error 02/52 and 02/54.
    """
    pass

class NoLabwareError(HamiltonStepError):
    """
    The labware to be loaded was not detected by autoload module.

    Note:

    May only occur on a Reload Carrier step if the labware property 'MlStarCarPosAreRecognizable' is set to 1.
    """
    pass

class UnexpectedLabwareError(HamiltonStepError):
    """
    The labware contains unexpected barcode ( may only occur on a Reload Carrier step ).
    """
    pass

class WrongLabwareError(HamiltonStepError):
    """
    The labware to be reloaded contains wrong barcode ( may only occur on a Reload Carrier step ).
    """
    pass

class BarcodeMaskError(HamiltonStepError):
    """
    The barcode read doesn't match the barcode mask defined.
    """
    pass

class BarcodeNotUniqueError(HamiltonStepError):
    """
    The barcode read is not unique. Previously loaded labware with same barcode was loaded without unique barcode check.
    """
    pass

class BarcodeAlreadyUsedError(HamiltonStepError):
    """
    The barcode read is already loaded as unique barcode ( it's not possible to load the same barcode twice ).
    """
    pass

class KitLotExpiredError(HamiltonStepError):
    """
    Kit Lot expired.
    """
    pass

class DelimiterError(HamiltonStepError):
    """
    Barcode contains character which is used as delimiter in result string.
    """
    pass


HAMILTON_ERROR_MAP = { 
    1: HamiltonSyntaxError,
    2: HardwareError,
    3: NotExecutedError,
    4: ClotError,
    5: BarcodeError,
    6: InsufficientLiquidError,
    7: TipPresentError,
    8: NoTipError,
    9: NoCarrierError,
    10: ExecutionError,
    11: PressureLLDError,
    12: CalibrateError,
    13: UnloadError,
    14: PressureLLDError,
    15: ParameterError,
    16: CoverOpenError,
    17: ImproperAspirationOrDispenseError,
    18: WashLiquidError,
    19: TemperatureError,
    20: TADMOvershotError,
    21: LabwareError,
    22: LabwareGrippedError,
    23: LabwareLostError,
    24: IllegalTargetPlatePositionError,
    25: IllegalInterventionError,
    26: TADMUndershotError,
    27: PositionError,
    28: UnexpectedcLLDError,
    29: AreaAlreadyOccupiedError,
    30: ImpossibleToOccupyAreaError,
    31: AntiDropControlError,
    32: DecapperError,
    33: DecapperHandlingError,
    99: SlaveError,
    100: WrongCarrierError,
    101: NoCarrierBarcodeError,
    102: LiquidLevelError,
    103: NotDetectedError,
    104: NotAspiratedError,
    105: ImproperDispensationError,
    106: NoLabwareError,
    107: UnexpectedLabwareError,
    108: WrongLabwareError,
    109: BarcodeMaskError,
    110: BarcodeNotUniqueError,
    111: BarcodeAlreadyUsedError,
    112: KitLotExpiredError,
    113: DelimiterError
}
"""
Maps integer error codes from Hamilton step return data to the appropriate `pyhamilton` errors
"""




"""
Pyhamilton
"""
import os
import shutil
from os.path import dirname, join, abspath
PACKAGE_PATH = abspath(dirname(__file__))
LAY_BACKUP_DIR = join(PACKAGE_PATH, 'LAY-BACKUP')
if not os.path.exists(LAY_BACKUP_DIR):
    os.mkdir(LAY_BACKUP_DIR)
OEM_STAR_PATH = join(PACKAGE_PATH, 'STAR-OEM')
if not (os.path.exists(OEM_STAR_PATH)
		and os.path.exists(os.path.join(OEM_STAR_PATH, 'RunHSLExecutor.dll'))
		and os.path.exists(os.path.join(OEM_STAR_PATH, 'HSLHttp.dll'))):
    raise FileNotFoundError('pyhamilton requires .../site-packages/pyhamilton/STAR-OEM, distributed separately.')
OEM_LAY_PATH = join(OEM_STAR_PATH, 'VENUS_Method', 'STAR_OEM_Test.lay')
OEM_HSL_PATH = join(OEM_STAR_PATH, 'VENUS_Method', 'STAR_OEM_noFan.hsl')
OEM_RUN_EXE_PATH = 'C:\\Program Files (x86)\\HAMILTON\\Bin\\HxRun.exe'
from .interface import *
from .deckresource import *
from .oemerr import *

from .utils import *

#from autoconfig import pyhamiltonconfig

this_file_dir = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.abspath(os.path.join(this_file_dir))
LIBRARY_DIR = os.path.join(PACKAGE_DIR, 'library')
TEMPLATE_DIR = os.path.join(PACKAGE_DIR, 'project-template')
exe_http = os.path.join(PACKAGE_DIR, 'bin', 'Hamilton HSLHttp Library Installer Version 2.7.exe')
exe_json = os.path.join(PACKAGE_DIR, 'bin', 'HSLJson Library v1.6 Installer.exe')


def full_paths_list(directory_abs_path):
    list_files = os.listdir(directory_abs_path)
    list_file_paths = [directory_abs_path + '\\' + file for file in list_files]
    return list_file_paths

def recursive_copy(source_dir, target_dir):
    source_list = full_paths_list(source_dir)
    for file in source_list:
        if os.path.isfile(file):
            shutil.copy(file, target_dir + '//' + os.path.basename(file))
        if os.path.isdir(file):
            target = target_dir + '//' + os.path.basename(file)
            print(target)
            if not os.path.exists(target):
                os.mkdir(target)
            recursive_copy(source_dir + '//' + os.path.basename(file), target)            


def autoconfig():
    print("Automatically configuring your PyHamilton installation")
    os.startfile(exe_http)
    os.startfile(exe_json)
    
    hamilton_lib_dir = os.path.abspath('C:/Program Files (x86)/HAMILTON/Library')
    print("Copying files to Hamilton library")
    print(LIBRARY_DIR)
    
    recursive_copy(LIBRARY_DIR, hamilton_lib_dir)        
    print("Configuration completed")

def create_project():
    current_dir = os.path.abspath(os.getcwd())
    print("Automatically configuring your PyHamilton installation")
    recursive_copy(TEMPLATE_DIR, current_dir)